import numpy as np


class Hist():
    def add_hist(self, name, xdata, **kwargs):
        """Plot a histogram.
        
        Args:
            name: data name.
            xdata: x-axis data.
            bins : int or sequence or str, default: 10
                If *bins* is an integer, it defines the number of equal-width bins
                in the range.

                If *bins* is a sequence, it defines the bin edges, including the
                left edge of the first bin and the right edge of the last bin;
                in this case, bins may be unequally spaced.  All but the last
                (righthand-most) bin is half-open.  In other words, if *bins* is::

                    [1, 2, 3, 4]

                then the first bin is ``[1, 2)`` (including 1, but excluding 2) and
                the second ``[2, 3)``.  The last bin, however, is ``[3, 4]``, which
                *includes* 4.

                If *bins* is a string, it is one of the binning strategies
                supported by `numpy.histogram_bin_edges`: 'auto', 'fd', 'doane',
                'scott', 'stone', 'rice', 'sturges', or 'sqrt'.

            range : tuple or None, default: None
                The lower and upper range of the bins. Lower and upper outliers
                are ignored. If not provided, *range* is ``(x.min(), x.max())``.
                Range has no effect if *bins* is a sequence.

                If *bins* is a sequence or *range* is specified, autoscaling
                is based on the specified bin range instead of the
                range of x.

            density : bool, default: False
                If ``True``, draw and return a probability density: each bin
                will display the bin's raw count divided by the total number of
                counts *and the bin width*
                (``density = counts / (sum(counts) * np.diff(bins))``),
                so that the area under the histogram integrates to 1
                (``np.sum(density * np.diff(bins)) == 1``).

                If *stacked* is also ``True``, the sum of the histograms is
                normalized to 1.

            weights : (n,) array-like or None, default: None
                An array of weights, of the same shape as *x*.  Each value in
                *x* only contributes its associated weight towards the bin count
                (instead of 1).  If *density* is ``True``, the weights are
                normalized, so that the integral of the density over the range
                remains 1.

                This parameter can be used to draw a histogram of data that has
                already been binned, e.g. using `numpy.histogram` (by treating each
                bin as a single point with a weight equal to its count) ::

                    counts, bins = np.histogram(data)
                    plt.hist(bins[:-1], bins, weights=counts)

                (or you may alternatively use `~.bar()`).

            cumulative : bool or -1, default: False
                If ``True``, then a histogram is computed where each bin gives the
                counts in that bin plus all bins for smaller values. The last bin
                gives the total number of datapoints.

                If *density* is also ``True`` then the histogram is normalized such
                that the last bin equals 1.

                If *cumulative* is a number less than 0 (e.g., -1), the direction
                of accumulation is reversed.  In this case, if *density* is also
                ``True``, then the histogram is normalized such that the first bin
                equals 1.

            bottom : array-like, scalar, or None, default: None
                Location of the bottom of each bin, ie. bins are drawn from
                ``bottom`` to ``bottom + hist(x, bins)`` If a scalar, the bottom
                of each bin is shifted by the same amount. If an array, each bin
                is shifted independently and the length of bottom must match the
                number of bins. If None, defaults to 0.

            histtype : {'bar', 'barstacked', 'step', 'stepfilled'}, default: 'bar'
                The type of histogram to draw.

                - 'bar' is a traditional bar-type histogram.  If multiple data
                  are given the bars are arranged side by side.
                - 'barstacked' is a bar-type histogram where multiple
                  data are stacked on top of each other.
                - 'step' generates a lineplot that is by default unfilled.
                - 'stepfilled' generates a lineplot that is by default filled.

            align : {'left', 'mid', 'right'}, default: 'mid'
                The horizontal alignment of the histogram bars.

                - 'left': bars are centered on the left bin edges.
                - 'mid': bars are centered between the bin edges.
                - 'right': bars are centered on the right bin edges.

            orientation : {'vertical', 'horizontal'}, default: 'vertical'
                If 'horizontal', `~.Axes.barh` will be used for bar-type histograms
                and the *bottom* kwarg will be the left edges.

            rwidth : float or None, default: None
                The relative width of the bars as a fraction of the bin width.  If
                ``None``, automatically compute the width.

                Ignored if *histtype* is 'step' or 'stepfilled'.

            log : bool, default: False
                If ``True``, the histogram axis will be set to a log scale.

            color : color or array-like of colors or None, default: None
                Color or sequence of colors, one per dataset.  Default (``None``)
                uses the standard line color sequence.

            label : str or None, default: None
                String, or sequence of strings to match multiple datasets.  Bar
                charts yield multiple patches per dataset, but only the first gets
                the label, so that `~.Axes.legend` will work as expected.

            stacked : bool, default: False
                If ``True``, multiple data are stacked on top of each other If
                ``False`` multiple data are arranged side by side if histtype is
                'bar' or on top of each other if histtype is 'step'
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]#tuple([round(np.random.uniform(0, 1),1) for _ in range(3)])
        elif isinstance(kwargs['color'], dict):
            kwargs['color'] = kwargs.pop('color')['mode']
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['plotmode'] = 'hist'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_hist
        return self
    
    def _execute_plot_hist(self, fig, ax, i, j):
        ax_plot = ax.hist(j['xdata'], **j['kwargs'])
    



from linora.image._image_io_aug import ImageIoAug
from linora.image._image_color_aug import ImageColorAug
from linora.image._image_noise_aug import ImageNoiseAug
from linora.image._image_resize_aug import ImageResizeAug
from linora.image._image_rescale_aug import ImageRescaleAug
from linora.image._image_position_aug import ImagePositionAug
from linora.image._image_crop_aug import ImageCropAug
from linora.image._image_filter_aug import ImageFilterAug
from linora.image._image_draw_aug import ImageDrawAug

__all__ = ['ImageAug']


class ImageAug(ImageIoAug, ImageColorAug, ImageNoiseAug, ImageResizeAug, 
               ImageRescaleAug, ImagePositionAug, ImageCropAug, 
               ImageFilterAug, ImageDrawAug
              ):
    """General class of image enhancement, which is used for pipeline image processing
    
    Args:
        image: a PIL instance.
        max_aug_nums: max image aug numbers, only valid for methods with parameter p.
        p: probability that the image does this. Default value is 1.
    """
    def __init__(self, image=None, max_aug_nums=None, p=1):
        super(ImageAug, self).__init__(image=image)
        if max_aug_nums is not None:
            if max_aug_nums>0:
                self._max_aug_nums = max_aug_nums
                self._nums = 1
            else:
                raise ValueError('`max_aug_nums` value error.')
        else:
            self._max_aug_nums = 0
        self._p = p



import time
import random
import functools
import itertools
from collections import defaultdict

import numpy as np
import pandas as pd

from linora.gfile._gfile import isfile
from linora.utils._config import Config


class DataSet():
    def __init__(self):
        self._params_init()
    
    def _params_init(self):
        self._params = Config()
        self._params.step = 1
        self._params.tensor = 'numpy'
        self._params.mode = 'total'
        self._params.mode1 = 'total'
        self._params.index_data = defaultdict()
        self._params.index_data['total'] = 'total'
        self._params.data_from = 'tensor'
        self._params.data = defaultdict()
        self._params.index = defaultdict(list)
        self._params.map = defaultdict(list)
        self._params.batch = defaultdict(list)
        self._params.batch[self._params.mode] = [0, False, 0]
        self._params.enumerate = defaultdict(int)
        self._params.options = defaultdict(dict)
        
    def batch(self, batch_size, drop_remainder=False):
        """Combines consecutive elements of this dataset into batches.
        
        Args:
            batch_size: representing the number of consecutive elements of this dataset to combine in a single batch.
            drop_remainder: representing whether the last batch should be dropped in the case it has fewer than batch_size elements; 
                            the default behavior is not to drop the smaller batch.
        """
        assert isinstance(batch_size, int) and batch_size>0, '`batch_size` type should be int and greater than 0.'
        self._params.batch[self._params.mode][0] = batch_size
        self._params.batch[self._params.mode][1] = drop_remainder
        self._params.options['batch'].update({self._params.step: {'batch_size':batch_size, 'drop_remainder':drop_remainder}})
        self._params.step += 1
        return self
    
    def cardinality(self):
        """Returns the cardinality of the dataset, if known."""
        return len(self._params.index[self._params.mode])
    
    def concatenate(self, datasets):
        """Creates a Dataset by concatenating the given dataset with this dataset.
        
        Args:
            datasets: la.data.Dataset or list of la.data.Dataset to be concatenated.
        """
        assert 'take_while' not in self._params.options, '`concatenate` must be placed in `take_while` front.'
        if not isinstance(datasets, list):
            self._concatenate(datasets)
        else:
            for dataset in datasets:
                assert self._params.data_mode==dataset._params.data_mode, 'The data types of the two data sets are inconsistent.'
            for dataset in datasets:
                self._concatenate(dataset)
        self._params.options['concatenate'].update({self._params.step: None})
        self._params.step += 1
        return self
    
    def _concatenate(self, dataset):
        if 'list' in self._params.data_mode:
            t = len(self._params.data[self._params.mode1][0])
        else:
            t = len(self._params.data[self._params.mode1])
        if 'list' in self._params.data_mode:
            assert len(self._params.data[self._params.mode1])==len(dataset._params.data[self._params.mode1]), 'Width needs to be consistent between data.'
            self._params.data[self._params.mode1] = [np.concatenate([self._params.data[self._params.mode1][i], dataset._params.data[self._params.mode1][i]]) for i in range(len(self._params.data[self._params.mode1]))]
        else:
            self._params.data[self._params.mode1] = np.concatenate([self._params.data[self._params.mode1], dataset._params.data[self._params.mode1]])
        self._params.index[self._params.mode] += [i+t for i in dataset._params.index[dataset._params.mode]]
        
    def drop(self, names):
        """Drop current dataset.
        
        Args:
            name: str or list, drop dataset name.
        """
        if isinstance(names, str):
            names = [names]
        for name in names:
            assert name!='total', "`names` can't be 'total'."
        for name in names:
            if name in self._params.index_data:
                self._params.index_data.pop(name)
            if name in self._params.data:
                if name in [j for i,j in self._params.index_data.items()]:
                    name1 = str(time.time()).split('.')[0]
                    self._params.data[name1] = self._params.data.pop(name)
                    for i,j in self._params.index_data.items():
                        if name==j:
                            self._params.index_data[i] = name1
                else:
                    self._params.data.pop(name)

            if name in self._params.index:
                self._params.index.pop(name)

            if name in self._params.map:
                self._params.map.pop(name)

            if name in self._params.batch:
                self._params.batch.pop(name)

            if name in self._params.enumerate:
                self._params.enumerate.pop(name)

            if self._params.mode==name:
                self._params.mode = 'total'
                self._params.mode1 = 'total'

            for i in list(self._params.data):
                if i not in [j for k,j in self._params.index_data.items()]:
                    self._params.data.pop(i)
        return self
        
    def enumerate(self, start=0):
        """Enumerates the elements of this dataset.
        
        Args:
            start: int, representing the start value for enumeration.
        """
        self._params.enumerate[self._params.mode] = start
        self._params.options['enumerate'].update({self._params.step: {'start':start}})
        self._params.step += 1
        return self
    
    def filter(self, filter_func):
        """A transformation that filter dataset based on a filter_func.
        
        Args:
            filter_func: A function that return True or False, datasets that are kept as True.
        """
        if self._params.data_mode=='list':
            filter_list = [i for i in range(len(self._params.data[self._params.mode1][0])) if filter_func([j[i] for j in self._params.data[self._params.mode1]])]
        else:
            filter_list = [r for r, i in enumerate(self._params.data[self._params.mode1]) if filter_func(i)]
        if filter_list:
            self._params.index[self._params.mode] = [i for i in self._params.index[self._params.mode] if i in filter_list]
        self._params.options['filter'].update({self._params.step: {'filter_func':filter_func}})
        self._params.step += 1
        return self
    
    def get(self, name):
        """Select current dataset.
        
        Args:
            name: split dataset name.
        """
        assert name in self._params.index, '`name` not in split dataset.'
        if self._params.batch[name][2]==-1:
            self._params.batch[name][0] = 0
        self._params.batch[name][2] = 0
        
        self._params.mode = name
        self._params.mode1 = self._params.index_data[name]
        
        for i in self._params.data:
            if i not in [j for k,j in self._params.index_data.items()]:
                self._params.data.pop(i)
        return self
    
    def join(self, join_dict, drop_exist_dataset=True):
        """Join Dataset.
        
        Args:
            join_dict: dict, {name: Dataset}, eg.{'train':la.data.Dataset.from_tensor()}.
            drop_exist_dataset: bool, If the name of the dataset is repeated, drop self exist dataset.
        """
        for name in join_dict:
            assert name!='total', "`name` can't be 'total'."
        for name in join_dict:
            if name in self._params.index:
                if drop_exist_dataset:
                    self.drop(name)
                    self._join(name, join_dict)
            else:
                self._join(name, join_dict)
        return self
    
    def _join(self, name, join_dict):
        self._params.data[name] = join_dict[name]._params.data[join_dict[name]._params.mode1].copy()
        self._params.index[name] = join_dict[name]._params.index[join_dict[name]._params.mode].copy()

        self._params.map[name] = join_dict[name]._params.map[join_dict[name]._params.mode].copy()
        self._params.batch[name] = join_dict[name]._params.batch[join_dict[name]._params.mode].copy()

        if join_dict[name]._params.mode in join_dict[name]._params.enumerate:
            self._params.enumerate[name] = join_dict[name]._params.enumerate[join_dict[name]._params.mode].copy()
            
        self._params.index_data[name] = name
            
    def list_names(self):
        """list datasets name."""
        return [i for i in self._params.index]
        
    def map(self, map_func, map_size=8):
        """Maps map_func across the elements of this dataset.
        
        Args:
            map_func: A function mapping a dataset element to another dataset element.
            map_size: representing the number elements to process asynchronously in parallel. 
        """
        assert isinstance(map_size, int) and map_size>0, '`map_size` type should be int and greater than 0.'
        self._params.map[self._params.mode] = [map_func, map_size]
        self._params.options['map'].update({self._params.step: {'map_func':map_func, 'map_size':map_size}})
        self._params.step += 1
        return self
    
    def options(self):
        """Returns the options for this dataset and its inputs."""
        return self._params.options
    
    def prefetch(self, prefetch_size):
        """Creates a Dataset that prefetches elements from this dataset.
        
        Args:
            prefetch_size: representing the maximum number of elements that will be buffered when prefetching.
        """
        assert 'take_while' not in self._params.options, '`prefetch` must be placed in `take_while` front.'
        assert isinstance(prefetch_size, int) and prefetch_size>0, '`prefetch_size` type should be int and greater than 0.'
        self._params.options['prefetch'].update({self._params.step: {'prefetch_size':prefetch_size}})
        self._params.step += 1
        return self

    def reduce(self, reduce_func):
        """Reduces the input dataset to a single element.
        
        Args:
            reduce_func: A function that maps to new_state. It must take two arguments and return a new element
        """
        if self._params.data_mode=='list':
            return [functools.reduce(reduce_func, i[self._params.index[self._params.mode]]) for i in self._params.data[self._params.mode1]]
        return functools.reduce(reduce_func, self._params.data[self._params.mode1][self._params.index[self._params.mode]])
    
    def rename(self, name_dict):
        """Rename current dataset.
        
        Args:
            name_dict: rename dataset name dict, eg.{'train':'train_set'}.
        """
        for name in name_dict:
            assert name!='total', "`name` can't be 'total'."
            assert name_dict[name]!='total', "`name` can't be 'total'."
            assert name in self._params.index, "name not exist."
            assert name_dict[name] not in self._params.index, "name already exist."
        for name in name_dict:
            if name in self._params.data:
                self._params.data[name_dict[name]] = self._params.data.pop(name)
                
            if name in self._params.index:
                self._params.index[name_dict[name]] = self._params.index.pop(name)
                
            if name in self._params.index_data:
                self._params.index_data[name_dict[name]] = self._params.index_data.pop(name)
            for i,j in self._params.index_data.items():
                if name==j:
                    self._params.index_data[i] = [name_dict[name]]
            
            if name in self._params.map:
                self._params.map[name_dict[name]] = self._params.map.pop(name)

            if name in self._params.batch:
                self._params.batch[name_dict[name]] = self._params.batch.pop(name)

            if name in self._params.enumerate:
                self._params.enumerate[name_dict[name]] = self._params.enumerate.pop(name)
                
            if self._params.mode==name:
                self._params.mode = name_dict[name]
                self._params.mode1 = self._params.index_data[self._params.mode]
        return self
    
    def repeat(self, repeat_size):
        """Repeats this dataset so each original value is seen count times.
        
        Args:
            repeat_size: representing the number of times the dataset should be repeated.
        """
        assert 'take_while' not in self._params.options, '`repeat` must be placed in `take_while` front.'
        assert isinstance(repeat_size, int) and repeat_size>0, '`repeat_size` type should be int and greater than 0.'
        self._params.index[self._params.mode] = self._params.index[self._params.mode]*(repeat_size+1)
        self._params.options['repeat'].update({self._params.step: {'repeat_size':repeat_size}})
        self._params.step += 1
        return self
    
    def sample(self, sample_size):
        """Sample data from Dataset.
        
        Args:
            sample_size: sample size.
        """
        loc = np.random.choice(self._params.index[self._params.mode], size=sample_size).tolist()
        if 'list' in self._params.data_mode:
            if self._params.mode in self._params.map:
                return self._to_tensor(self._batch_list_map(loc))
            else:
                return self._to_tensor(self._batch_list(loc))
        elif self._params.mode in self._params.map:
            return self._to_tensor(self._batch_map(loc))
        else:
            return self._to_tensor(self._batch(loc))
    
    def shard(self, shard_size, shard_index):
        """Creates a Dataset that includes only 1/num_shards of this dataset.
        
        Args:
            shard_size: representing the number of shards operating in parallel.
            shard_index: representing the worker index.
        """
        assert 'take_while' not in self._params.options, '`shard` must be placed in `take_while` front.'
        assert isinstance(shard_size, int) and shard_size>0, '`shard_size` type should be int and greater than 0.'
        assert isinstance(shard_index, int) and shard_index>=0, '`shard_index` type should be int and greater than or equal to 0.'
        self._params.index[self._params.mode] = [self._params.index[self._params.mode][i] for i in range(shard_index, len(self._params.index[self._params.mode]), shard_size)]
        self._params.options['shard'].update({self._params.step: {'shard_size':shard_size, 'shard_index':shard_index}})
        self._params.step += 1
        return self
    
    def shuffle(self, shuffle_size, seed=None):
        """Randomly shuffles the elements of this dataset.
        
        Args:
            shuffle_size: representing the number of elements from this dataset from which the new dataset will sample.
            seed: representing the random seed that will be used to create the distribution.
        """
        assert 'take_while' not in self._params.options, '`shuffle` must be placed in `take_while` front.'
        assert isinstance(shuffle_size, int) and shuffle_size>-2 and shuffle_size!=0, '`shuffle_size` type should be int and greater than 0 or equal to -1.'
        if isinstance(self._params.index[self._params.mode], list):
            self._params.index[self._params.mode] = pd.Series(index=self._params.index[self._params.mode], data=1).index
        if shuffle_size > 0:
            t = [self._params.index[self._params.mode][shuffle_size*i:shuffle_size*(i+1)].to_list() for i in range(len(self._params.index[self._params.mode])//shuffle_size+1)]
            [random.shuffle(i, random=lambda :((seed if seed is not None else random.randint(1, 99))+self._params.batch[self._params.mode][2])%10/10) for i in t]
            self._params.index[self._params.mode] = list(itertools.chain.from_iterable(t))
        else:
            self._params.index[self._params.mode] = self._params.index[self._params.mode].to_series().sample(frac=1, random_state=seed).tolist()
        self._params.options['shuffle'].update({self._params.step: {'shuffle_size':shuffle_size, 'seed':seed}})
        self._params.step += 1
        return self
    
    def skip(self, skip_size):
        """Creates a Dataset that skips count elements from this dataset.
        
        Args:
            skip_size: representing the number of elements of this dataset that should be skipped to form the new dataset. 
                       If count is greater than the size of this dataset, the new dataset will contain no elements.
        """
        assert 'take_while' not in self._params.options, '`skip` must be placed in `take_while` front.'
        assert isinstance(skip_size, int) and skip_size>0, '`skip_size` type should be int and greater than 0.'
        self._params.index[self._params.mode] = self._params.index[self._params.mode][skip_size:]
        self._params.options['skip'].update({self._params.step: {'skip_size':skip_size}})
        self._params.step += 1
        return self
    
    def split(self, split_dict, shuffle=True, seed=None):
        """Split Dataset.
        
        Args:
            split_dict: dict, {data_name:data_rate}, eg.{'train':0.7, 'test':0.3}.
            shuffle: whether randomly shuffles the elements of this dataset.
            seed: random seed.
        """
        for i in split_dict:
            assert i not in self._params.index, f"`{i}` has exist."
            assert i!='total', "`split_dict` key can't be 'total'."
        t = sum(split_dict[i] for i in split_dict)
        t = {i:split_dict[i]/t for i in split_dict}
        if self._params.data_from in ['from_folder', 'from_class_folder']:
            if isinstance(self._params.data[self._params.mode1], list):
                label = self._params.data[self._params.mode1][1][self._params.index[self._params.mode]]
                index = np.array(self._params.index[self._params.mode])
                for i in np.unique(label):
                    index1 = index[label==i].tolist()
                    n = 0
                    for j in t:
                        self._params.index[j] += index1[n:n+int(t[j]*len(index1))]
                        n += int(t[j]*len(index1))
                if shuffle:
                    for i in t:
                        self._params.index[i] = pd.Series(self._params.index[i]).sample(frac=1, random_state=seed).tolist()
            else:
                self._split(t, shuffle, seed)
        else:
            self._split(t, shuffle, seed)
        
        for i in split_dict:
            self._params.batch[i] = [0, False, 0]
            self._params.index_data[i] = self._params.mode1
        return self
    
    def _split(self, t, shuffle, seed):
        if shuffle:
            index = pd.Series(self._params.index[self._params.mode]).sample(frac=1, random_state=seed).tolist()
        else:
            index = self._params.index[self._params.mode]
        n = 0
        for i in t:
            self._params.index[i] += index[n:n+int(t[i]*len(index))]
            n += int(t[i]*len(index))
        
    def take(self, take_size):
        """Creates a Dataset with at most count elements from this dataset.
        
        Args:
            take_size: representing the number of elements of this dataset that should be taken to form the new dataset. 
                       If count is -1, or if count is greater than the size of this dataset, 
                       the new dataset will contain all elements of this dataset.
        """
        assert 'take_while' not in self._params.options, '`take` must be placed in `take_while` front.'
        assert isinstance(take_size, int) and take_size>-2 and take_size!=0, '`take_size` type should be int and greater than 0 or equal to -1.'
        if take_size != -1:
            self._params.index[self._params.mode] = self._params.index[self._params.mode][:take_size]
        self._params.options['take'].update({self._params.step: {'take_size':take_size}})
        self._params.step += 1
        return self
    
    def take_while(self, take_func):
        """A transformation that stops dataset iteration based on a take_func.
        
        Args:
            take_func: A function that return True or False
        """
        temp = set()
        index = self._params.index[self._params.mode][:max([self._params.index[self._params.mode].index(i) for i in range(len(self._params.data[self._params.mode1]))])+1]
        for r, i in enumerate(index):
            if i in temp:
                continue
            temp.add(i)
            if 'list' in self._params.data_mode:
                if take_func([j[i] for j in self._params.data[self._params.mode1]]):
                    self._params.index[self._params.mode] = self._params.index[self._params.mode][:r]
                    break
            else:
                if take_func(self._params.data[self._params.mode1][i]):
                    self._params.index[self._params.mode] = self._params.index[self._params.mode][:r]
                    break
        self._params.options['take_while'].update({self._params.step: {'take_func':take_func}})
        self._params.step += 1
        return self
    
    def to_tensor(self, mode='tf'):
        """Transform data from numpy array to tensor.
        
        Args:
            mode: Deep learning framework name, one of ['tf', 'pytorch', 'paddle', 'mxnet', 'mindspore'].
        """
        assert 'take_while' not in self._params.options, '`to_tensor` must be placed in `take_while` front.'
        if self._params.tensor==mode:
            return self
        if mode in ['tf', 'tensorflow']:
            if 'tf' in globals():
                self._params.framework = tf.convert_to_tensor
            else:
                from tensorflow import convert_to_tensor
                self._params.framework = convert_to_tensor
        elif mode in ['pytorch', 'torch']:
            if 'torch' in globals():
                self._params.framework = torch.as_tensor
            else:
                from torch import as_tensor
                self._params.framework = as_tensor
        elif mode in ['paddle', 'paddlepaddle']:
            if 'paddle' in globals():
                self._params.framework = paddle.to_tensor
            else:
                from paddle import to_tensor
                self._params.framework = to_tensor
        elif mode in ['mx', 'mxnet']:
            if 'mx' in globals():
                self._params.framework = mx.ndarray.array
            elif 'mxnet' in globals():
                self._params.framework = mxnet.ndarray.array
            else:
                from mxnet.ndarray import array
                self._params.framework = array
        elif mode in ['mindspore']:
            from mindspore.numpy import array
            self._params.framework = array
        else:
            raise ValueError('`mode` value error.')
        self._params.tensor = mode
        self._params.options['to_tensor'].update({self._params.step: {'mode':mode}})
        self._params.step += 1
        return self
    
    def unbatch(self):
        """Splits elements of a dataset into multiple elements."""
        assert not isinstance(self._params.data[self._params.mode1], list), 'Input data cannot be a tuple.'
        assert self._params.mode=='total', f'{self._params.mode} dataset not supported.'
        self._params.data[self._params.mode1] = np.array(list(itertools.chain.from_iterable(self._params.data[self._params.mode1])))
        self._params.index[self._params.mode] = list(range(len(self._params.data[self._params.mode1])))
        return self
    
    def unique(self):
        """A transformation that discards duplicate elements of a Dataset."""
        if isinstance(self._params.data[self._params.mode1], list):
            return tuple([np.unique(i) for i in self._params.data[self._params.mode1][self._params.index[self._params.mode]]])
        else:
            return np.unique(self._params.data[self._params.mode1][self._params.index[self._params.mode]])
    
    def _to_tensor(self, data):
        if self._params.tensor=='numpy':
            return data
        if 'list' in self._params.data_mode:
            return [self._params.framework(i) for i in data]
        else:
            return self._params.framework(data)
    
    def _data_mode(self):
        self._params.data_mode = 'list_array' if isinstance(self._params.data[self._params.mode1], list) else 'array'
        if isinstance(self._params.data[self._params.mode1], list):
            t = [i[0] for i in self._params.data[self._params.mode1]]
        else:
            t = self._params.data[self._params.mode1][0]
        if isinstance(t, str):
            if isfile(t):
                if t.split('.')[-1] in ['png', 'jpg', 'jpeg', 'bmp', 'rgb', 'tif', 'tiff', 'webp']:
                    self._params.data_mode = 'image'
        elif isinstance(t, list):
            for i in t:
                if isinstance(i, str):
                    if isfile(i):
                        if i.split('.')[-1] in ['png', 'jpg', 'jpeg', 'bmp', 'rgb', 'tif', 'tiff', 'webp']:
                            self._params.data_mode = 'list_image'
    
    def __iter__(self):
        if 'list' in self._params.data_mode:
            if self._params.mode in self._params.map:
                self._batch_func = self._batch_list_map
            else:
                self._batch_func = self._batch_list
        elif self._params.mode in self._params.map:
            self._batch_func = self._batch_map
        else:
            self._batch_func = self._batch
        return self
    
    def __next__(self):
        if self._params.batch[self._params.mode][0]==0:
            self._params.batch[self._params.mode][0] = 1
            self._params.batch[self._params.mode][2] = -1
            if self._params.mode in self._params.enumerate:
                self._params.enumerate[self._params.mode] += 1
                return (self._params.enumerate[self._params.mode]-1, self._to_tensor(self._batch_func(self._params.index[self._params.mode])))
            return self._to_tensor(self._batch_func(self._params.index[self._params.mode]))
        loc = self._params.index[self._params.mode][self._params.batch[self._params.mode][0]*self._params.batch[self._params.mode][2]:self._params.batch[self._params.mode][0]*(self._params.batch[self._params.mode][2]+1)]
        if len(loc)==0:
            raise StopIteration
        elif len(loc)<self._params.batch[self._params.mode][0]:
            if self._params.batch[self._params.mode][1]:
                raise StopIteration
        self._params.batch[self._params.mode][2] += 1
        if self._params.mode in self._params.enumerate:
            self._params.enumerate[self._params.mode] += 1
            return (self._params.enumerate[self._params.mode]-1, self._to_tensor(self._batch_func(loc)))
        return self._to_tensor(self._batch_func(loc))


import time

from linora.utils._config import Config
from linora.utils._logger import Logger

__all__ = ['Progbar']


class Progbar():
    """Displays a progress bar."""
    def __init__(self, target, width=25, verbose=1, unit_name='step'):
        """
        Args:
            target: Total number of steps expected, None if unknown.
            width: Progress bar width on screen.
            verbose: Verbosity mode, 0 (silent), 1 (verbose)
            unit_name: Display name for step counts (usually "step" or "sample").
        """
        self.param = Config()
        self.param.width = width
        self.param.target = target
        self.param.time = time.time()
        self.param.n = 0
        self.param.unit_name = unit_name
        self.param.verbose = verbose
        self.param.current = 0
        if verbose:
            self.param.logger = Logger()

    def add(self, current, values=None):
        """
        Args:
            current: add Index of current step, current += current.
            values: List of tuples: (name, value_for_last_step). 
        """
        if not self.param.verbose:
            return 0
        self.param.n += 1
        self.param.current += current
        if self.param.target is not None:
            if self.param.current>self.param.target:
                self.param.current = self.param.target
            rate = self.param.current/self.param.target
            percent = int(rate*self.param.width)
            msg = f"{self.param.current}/{self.param.target} "
            msg = msg+f"[{('='*percent+'>'+'.'*(self.param.width-percent))[:self.param.width]}] "
        else:
            msg = f"{self.param.current}/Unknown "
        
        time_diff = time.time()-self.param.time
        if self.param.target is not None:
            if self.param.current<self.param.target:
                msg = msg+f"- {rate*100:.1f}% EAT: {int(time_diff/self.param.current*(self.param.target-self.param.current))}s"
            else:
                msg = msg+f"- {int(time_diff/self.param.n*1000)}ms/{self.param.unit_name}"
        else:
            msg = msg+f"- {int(time_diff/self.param.n*1000)}ms/{self.param.unit_name}"
        if values is not None:
            msg = msg+' - '+''.join([f"{i[0]}: {i[1]} " for i in values])
        
        if self.param.target is None:
            self.param.logger.info(msg+' '*4, enter=False)
        elif self.param.current<self.param.target:
            self.param.logger.info(msg+' '*4, enter=False)
        else:
            self.param.logger.info(msg+' '*4, enter=True)
            
    def update(self, current, values=None):
        """
        Args:
            current: update Index of current step.
            values: List of tuples: (name, value_for_last_step). 
        """
        if not self.param.verbose:
            return 0
        self.param.n += 1
        if self.param.target is not None:
            if current>self.param.target:
                raise
            rate = current/self.param.target
            percent = int(rate*self.param.width)
            msg = f"{current}/{self.param.target} "
            msg = msg+f"[{('='*percent+'>'+'.'*(self.param.width-percent))[:self.param.width]}] "
        else:
            msg = f"{current}/Unknown "
        
        time_diff = time.time()-self.param.time
        if self.param.target is not None:
            if current<self.param.target:
                msg = msg+f"- {rate*100:.1f}% EAT: {int(time_diff/current*(self.param.target-current))}s"
            else:
                msg = msg+f"- {int(time_diff/self.param.n*1000)}ms/{self.param.unit_name}"
        else:
            msg = msg+f"- {int(time_diff/self.param.n*1000)}ms/{self.param.unit_name}"
        if values is not None:
            msg = msg+' - '+''.join([f"{i[0]}: {i[1]} " for i in values])
        
        if self.param.target is None:
            self.param.logger.info(msg+' '*4, enter=False)
        elif current<self.param.target:
            self.param.logger.info(msg+' '*4, enter=False)
        else:
            self.param.logger.info(msg+' '*4, enter=True)



import numpy as np

__all__ = ['timeseries_train_test_split', 'timeseries_walk_forward_fold', 'timeseries_kfold']


def timeseries_train_test_split(df, test_size=0.2, gap=0, ascending=True):
    """Split DataFrame or matrices into random train and test subsets for timeseries
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        test_size: float, optional (default=0.2)
            Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split.
        gap: int, default=0
            Represents the absolute number of the dropped samples between training set and test set.
        ascending  : boolean, default=True, optional, Whether timeseries is ascending.
    Returns:
        list, length=2, List each list containing train-test split of inputs.
    """
    assert gap>=0, "`gap` should be great than or equal to 0."
    t = df.index.tolist()
    if not ascending:
        t.reverse()
    train_len = round(len(t)*(1-test_size))-gap
    return [t[0:train_len], t[train_len+gap:]]


def timeseries_walk_forward_fold(df, n_splits=3, gap=0, ascending=True):
    """Walk Forward Folds cross-validator for timeseries
    
    Provides train/test indices to split data in train/test sets. 
    Split dataset into k consecutive folds.
    Each fold train set length is not equal.
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        n_splits : int, default=3
            Number of folds. Must be at least 2.
        gap: int, default=0
            Represents the absolute number of the dropped samples between training set and test set.
        ascending  : boolean, default=True, optional, Whether timeseries is ascending.
    Returns:
        list, length=n_splits, List each list containing train-test split of inputs.
    """
    assert gap>=0, "`gap` should be great than or equal to 0."
    t = df.index.tolist()
    if not ascending:
        t.reverse()
    t_len = len(t)
    m = int(np.floor(t_len/(n_splits+1)))
    return [[t[:t_len-m*(i+1)-gap], t[t_len-m*(i+1):t_len-m*i]] for i in range(n_splits)]


def timeseries_kfold(df, train_size=0.3, test_size=0.1, gap=0, ascending=True):
    """K-Folds cross-validator for timeseries
    
    Provides train/test indices to split data in train/test sets. 
    Split dataset into k consecutive folds.
    Each fold train set length is equal.
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        train_size: float, optional (default=0.3)
            Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the train split.
        test_size: float, optional (default=0.1)
            Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split.
        gap: int, default=0
            Represents the absolute number of the dropped samples between training set and test set.
        ascending  : boolean, default=True, optional, Whether timeseries is ascending.
    
    Returns:
        list, length=`int(np.floor((len(df)-len(trainset)-gap)/len(testset)))`
        List each list containing train-test split of inputs.
    """
    assert gap>=0, "`gap` should be great than or equal to 0."
    assert train_size+test_size<1, "`train_size`+`test_size` should be less than 1."
    t = df.index.tolist()
    if not ascending:
        t.reverse()
    t_len = len(t)
    train_len = int(np.floor(t_len*train_size))
    test_len = int(np.floor(t_len*test_size))
    all_len = train_len+test_len+gap
    n_splits = int(np.floor((t_len-all_len)/test_len)+1)
    return [[t[test_len*i:test_len*i+train_len], t[test_len*i+train_len+gap:test_len*i+all_len]] for i in range(n_splits)]



import time
from collections import defaultdict

from linora.gfile._gfile import exists, remove
from linora.utils._config import Config

__all__ = ['CSVLogger']


class CSVLogger():
    """Callback that streams epoch results to a CSV file.
    
    Args:
        filename: Filename of the CSV file, e.g. 'run/log.csv'.
        sep: String used to separate elements in the CSV file.
        append: Boolean. True: append if file exists (useful for continuing training). False: overwrite existing file.
        wait_num: int, default 10, How many batches to store at intervals.
    """
    def __init__(self, filename, sep=',', append=False, wait_num=10):
        self._params = Config()
        self._params.filename = filename
        self._params.sep = sep
        self._params.append = append
        self._params.wait_num = wait_num
        self._params.history = defaultdict()
        if exists(filename):
            if not append:
                remove(filename)
        self._params.polt_num = 0
        self._params.name = 'CSVLogger'
        
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        self._params.history[self._params.polt_num] = [time.time(), batch, log]
        if self._params.polt_num%self._params.wait_num==0:
            with open(self._params.filename, 'a+') as f:
                for i,j in self._params.history.items():
                    msg = self._params.sep.join([str(r)+':'+str(j[2][r]) for r in j[2]])
                    f.write(f'time:{j[0]}{self._params.sep}batch:{j[1]}{self._params.sep}{msg}\n')
            self._params.history = defaultdict()
        self._params.polt_num += 1
        


from multiprocessing import cpu_count

import numpy as np

from linora.param_search._HyperParameters import HyperParametersRandom, HyperParametersGrid

__xgboost_version__ = '1.4.0'
__lightgbm_version__ = '3.3.0'
n_jobs = int(np.ceil(cpu_count()*0.9))

def model_hp(model, method):
    if method=='random':
        hp = HyperParametersRandom()
    elif method=='grid':
        hp = HyperParametersGrid()
    else:
        raise ValueError('`method` value error.')
    if model=='XGBClassifier' and method=='random':
        hp.Float('learning_rate', 0.01, 0.1)
        hp.Int('n_estimators', 100, 850)
        hp.Choice('max_depth', [3, 4, 5, 6, 7])
        hp.Choice('min_child_weight', [1, 2, 3, 4, 5, 6, 7])
        hp.Choice('max_delta_step', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('subsample', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bytree', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bylevel', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bynode', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('gamma', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('random_state', [27])
        hp.Choice('booster', ['gbtree'])
        hp.Choice('importance_type', ["gain", "weight", "cover", "total_gain", "total_cover"])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['binary:logistic'])
        hp.Choice('tree_method', ['auto', 'exact', 'approx', 'hist'])
        hp.Choice('gpu_id', [-1])
    elif model=='XGBRegressor' and method=='random':
        hp.Float('learning_rate', 0.01, 0.1)
        hp.Int('n_estimators', 100, 850)
        hp.Choice('max_depth', [3, 4, 5, 6, 7])
        hp.Choice('min_child_weight', [1, 2, 3, 4, 5, 6, 7])
        hp.Choice('max_delta_step', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('subsample', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bytree', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bylevel', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bynode', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
        hp.Choice('gamma', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('scale_pos_weight', [1])
        hp.Choice('random_state', [27])
        hp.Choice('booster', ['gbtree'])
        hp.Choice('importance_type', ["gain", "weight", "cover", "total_gain", "total_cover"])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['reg:squarederror'])
        hp.Choice('tree_method', ['auto', 'exact', 'approx', 'hist'])
        hp.Choice('gpu_id', [-1])
    elif model=='XGBClassifier' and method=='grid':
        hp.Choice('learning_rate', [0.01, 0.03, 0.05, 0.07, 0.09, 0.1], 0.1, rank=7)
        hp.Choice('n_estimators', list(range(100, 850, 50)), 300, rank=1)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], 5, rank=2)
        hp.Choice('min_child_weight', [1, 2, 3, 4, 5], 1, rank=2)
        hp.Choice('max_delta_step', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 0, rank=6)
        hp.Choice('reg_alpha', [0, 0.05, 0.5, 1, 50, 100], 0, rank=5)
        hp.Choice('reg_lambda', [0.05, 0.5, 1, 10, 50, 100], 1, rank=5)
        hp.Choice('subsample', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bytree', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bylevel', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bynode', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('gamma', [0.1, 0.2, 0.3, 0.4, 0.5], 0, rank=3)
        hp.Choice('random_state', [27])
        hp.Choice('booster', ['gbtree'])
        hp.Choice('importance_type', ["gain"])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['binary:logistic'])
        hp.Choice('tree_method', ['hist'])
        hp.Choice('gpu_id', [-1])
    elif model=='XGBRegressor' and method=='grid':
        hp.Choice('learning_rate', [0.01, 0.03, 0.05, 0.07, 0.09, 0.1], 0.1, rank=7)
        hp.Choice('n_estimators', list(range(100, 850, 50)), 300, rank=1)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], 5, rank=2)
        hp.Choice('min_child_weight', [1, 2, 3, 4, 5], 1, rank=2)
        hp.Choice('max_delta_step', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 0, rank=6)
        hp.Choice('reg_alpha', [0, 0.05, 0.5, 1, 50, 100], 0, rank=5)
        hp.Choice('reg_lambda', [0.05, 0.5, 1, 10, 50, 100], 1, rank=5)
        hp.Choice('subsample', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bytree', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bylevel', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('colsample_bynode', [0.6, 0.7, 0.8, 0.9], 0.8, rank=4)
        hp.Choice('gamma', [0.1, 0.2, 0.3, 0.4, 0.5], 0, rank=3)
        hp.Choice('scale_pos_weight', [1])
        hp.Choice('random_state', [27])
        hp.Choice('booster', ['gbtree'])
        hp.Choice('importance_type', ["gain"])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['reg:squarederror'])
        hp.Choice('tree_method', ['hist'])
        hp.Choice('gpu_id', [-1])
    elif model=='LGBMClassifier' and method=='grid':
        hp.Choice('learning_rate', [0.01, 0.03, 0.05, 0.07, 0.09, 0.1], 0.1, rank=5)
        hp.Choice('n_estimators', list(range(100, 850, 50)), 300, rank=1)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], 4, rank=2)
        hp.Dependence('num_leaves', 'max_depth', lambda x:int(np.random.choice(range(int(2**x*0.4), int(2**x*0.6)))), default=15)
        hp.Choice('reg_alpha', [0, 0.05, 0.5, 1, 50, 100], 0, rank=4)
        hp.Choice('reg_lambda', [0.05, 0.5, 1, 10, 50, 100], 1, rank=4)
        hp.Choice('subsample', [0.6, 0.7, 0.8, 0.9], 0.8, rank=3)
        hp.Choice('colsample_bytree', [0.6, 0.7, 0.8, 0.9], 0.8, rank=3)
        hp.Choice('random_state', [27])
        hp.Choice('boosting_type', ['gbdt'])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['binary'])
    elif model=='LGBMClassifier' and method=='random':
        hp.Float('learning_rate', 0.01, 0.1, default=0.1)
        hp.Int('n_estimators', 100, 850, default=200)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], default=4)
        hp.Dependence('num_leaves', 'max_depth', lambda x:int(np.random.choice(range(int(2**x*0.4), int(2**x*0.6)))), default=15)
        hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('subsample', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bytree', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('random_state', [27])
        hp.Choice('boosting_type', ['gbdt'])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['binary'])
    elif model=='LGBMRegressor' and method=='grid':
        hp.Choice('learning_rate', [0.01, 0.03, 0.05, 0.07, 0.09, 0.1], 0.1, rank=5)
        hp.Choice('n_estimators', list(range(100, 850, 50)), 300, rank=1)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], 4, rank=2)
        hp.Dependence('num_leaves', 'max_depth', lambda x:int(np.random.choice(range(int(2**x*0.4), int(2**x*0.6)))), default=15)
        hp.Choice('reg_alpha', [0, 0.05, 0.5, 1, 50, 100], 0, rank=4)
        hp.Choice('reg_lambda', [0.05, 0.5, 1, 10, 50, 100], 1, rank=4)
        hp.Choice('subsample', [0.6, 0.7, 0.8, 0.9], 0.8, rank=3)
        hp.Choice('colsample_bytree', [0.6, 0.7, 0.8, 0.9], 0.8, rank=3)
        hp.Choice('random_state', [27])
        hp.Choice('boosting_type', ['gbdt'])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['regression'])
    elif model=='LGBMRegressor' and method=='random':
        hp.Float('learning_rate', 0.01, 0.1, default=0.1)
        hp.Int('n_estimators', 100, 850, default=200)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], default=4)
        hp.Dependence('num_leaves', 'max_depth', lambda x:int(np.random.choice(range(int(2**x*0.4), int(2**x*0.6)))), default=15)
        hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('subsample', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bytree', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('random_state', [27])
        hp.Choice('boosting_type', ['gbdt'])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [n_jobs])
        hp.Choice('objective', ['regression'])
    return hp


import random
import itertools
from collections import defaultdict

import numpy as np

__all__ = ['HyperParametersRandom', 'HyperParametersGrid']


class HyperParametersRandom():
    def __init__(self):
        """Container for both a hyperparameter space, and current values."""
        self._space = defaultdict()
        self.params = defaultdict()
        self.params_history = defaultdict()
        self._update_init = True
        self._update_nums = 0
        
    def Boolean(self, name, default=None):
        """Choice between True and False.

        Args:
            name: Str. Name of parameter. Must be unique.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Boolean', 'default':default}
            self.params[name] = np.random.choice([True, False]) if default is None else default
        
    def Choice(self, name, values, weight=None, default=None):
        """Choice of one value among a predefined set of possible values.

        Args:
            name: Str. Name of parameter. Must be unique.
            values: List of possible values. Values must be int, float, str, or bool. All values must be of the same type.
            weight: List of possible values, The probabilities associated with each entry in value.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Choice', 'values':list(values), 'weight':weight, 'default':default}
            self.params[name] = np.random.choice(self._space[name]['values'], p=weight) if default is None else default
    
    def Dependence(self, name, dependent_name, function, default=None):
        """Values generated depending on other parameters.

        Args:
            name: Str. Name of parameter. Must be unique.
            dependent_name: str, dependent params name, must already exist.
            function: A function to transform the value of `dependent_name`.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        assert dependent_name in self._space, '`dependent_name` must already exist.'
        if name not in self._space:
            self._space[name] = {'mode':'Dependence', 'dependent_name':dependent_name, 'function':function, 'default':default}
            self.params[name] = function(self.params[dependent_name]) if default is None else default
    
    def Fixed(self, name, value):
        """Fixed, untunable value.
        
        Args:
            name: Str. Name of parameter. Must be unique.
            value: Fixed value.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Fixed', 'default':value}
            self.params[name] = value
    
    def Float(self, name, min_value, max_value, rounds=2, default=None):
        """Floating point range, can be evenly divided.

        Args:
            name: Str. Name of parameter. Must be unique.
            min_value: Float. Lower bound of the range.
            max_value: Float. Upper bound of the range.
            rounds: Optional. Int, e.g. 2 mean round(x, 2). smallest meaningful distance between two values. 
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Float', 'min_value':min_value, 'max_value':max_value, 'round':rounds, 'default':default}
            self.params[name] = round(random.uniform(min_value, max_value), rounds) if default is None else default
    
    def Int(self, name, min_value, max_value, default=None):
        """Integer range.
        Note that unlinke Python's range function, max_value is included in the possible values this parameter can take on.

        Args:
            name: Str. Name of parameter. Must be unique.
            min_value: Int. Lower limit of range (included).
            max_value: Int. Upper limit of range (included).
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Int', 'min_value':min_value, 'max_value':max_value+1, 'default':default}
            self.params[name] = np.random.randint(min_value, max_value+1) if default is None else default
        
    def update(self, best_params=None):
        """params update"""
        if self._update_init:
            self._update_init = False
        else:
            for name, config in self._space.items():
                if config['mode']=='Boolean':
                    self.params[name] = np.random.choice([True, False])
                elif config['mode']=='Float':
                    self.params[name] = round(np.random.uniform(config['min_value'], config['max_value']), config['round'])
                elif config['mode']=='Int':
                    self.params[name] = np.random.randint(config['min_value'], config['max_value'])
                elif config['mode']=='Choice':
                    self.params[name] = np.random.choice(config['values'], p=config['weight'])
                elif config['mode']=='Dependence':
                    self.params[name] = config['function'](self.params[config['dependent_name']])
        self.params_history[self._update_nums] = self.params.copy()
        self._update_nums += 1
        if best_params is not None:
            self.best_params = best_params.copy()

    def from_HyperParameters(self, hp):
        """update HyperParametersRandom class.
        
        Args:
            hp: a HyperParametersRandom class.
        """
        for name, config in hp._space.items():
            self._space[name] = config
            self.params[name] = hp.params[name]

class HyperParametersGrid():
    def __init__(self):
        """Container for both a hyperparameter space, and current values."""
        self._space = defaultdict()
        self.params = defaultdict()
        self.params_history = defaultdict()
        self._update_init = True
        self._update_nums = 0
        self._rank_list = []
        self._rank = defaultdict(list)
        self._dependent = defaultdict(list)
        
        
    def Boolean(self, name, default=None, rank=0):
        """Choice between True and False.

        Args:
            name: Str. Name of parameter. Must be unique.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
            rank: Int, default 0, Importance ordering of parameters, smaller is more important, rank should be greater than 1.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Boolean', 'default':default, 'rank':rank, 'values':[True, False]}
            if rank>0:
                self._rank[rank].append(name)
            self.params[name] = np.random.choice([True, False]) if default is None else default
        
    def Choice(self, name, values, default=None, rank=0):
        """
        Choice of one value among a predefined set of possible values.

        Args:
            name: Str. Name of parameter. Must be unique.
            values: List of possible values. Values must be int, float, str, or bool. All values must be of the same type.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
            rank: Int, default 0, Importance ordering of parameters, smaller is more important, rank should be greater than 1.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Choice', 'default':default, 'rank':rank, 'values':list(values)}
            if rank>0:
                self._rank[rank].append(name)
            self.params[name] = random.choice(self._space[name]['values']) if default is None else default
        
    def Dependence(self, name, dependent_name, function, default=None):
        """Values generated depending on other parameters.

        Args:
            name: Str. Name of parameter. Must be unique.
            dependent_name: str, dependent params name, must already exist.
            function: A function to transform the value of `dependent_name`.
            default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        assert dependent_name in self._space, '`dependent_name` must already exist.'
        if name not in self._space:
            self._space[name] = {'mode':'Dependence', 'default':default,
                                 'dependent_name':dependent_name, 'function':function}
            self._dependent[dependent_name].append(name)
            self.params[name] = function(self.params[dependent_name]) if default is None else default
    
    def Fixed(self, name, value):
        """Fixed, untunable value.
        
        Args:
            name: Str. Name of parameter. Must be unique.
            value: Fixed value.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Fixed', 'default':value}
            self.params[name] = value
        
    def Float(self, name, min_value, max_value, rounds=1, default=None, rank=0):
        """Floating point range, can be evenly divided.

        Args:
            name: Str. Name of parameter. Must be unique.
            min_value: Float. Lower bound of the range.
            max_value: Float. Upper bound of the range.
            rounds: Optional. Int, e.g. 2 mean round(x, 2). smallest meaningful distance between two values. 
            default: Default value to return for the parameter. If unspecified, the default value will be None.
            rank: Int, default 0, Importance ordering of parameters, smaller is more important, rank should be greater than 1.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Float', 'default':default, 'rank':rank, 
                                 'values':np.linspace(min_value, max_value, round((max_value-min_value)*10**rounds)+1).round(rounds).tolist()}
            if rank>0:
                self._rank[rank].append(name)
            self.params[name] = round(random.uniform(min_value, max_value), rounds) if default is None else default
        
    def Int(self, name, min_value, max_value, default=None, rank=0):
        """Integer range.
        Note that unlinke Python's range function, max_value is included in the possible values this parameter can take on.

        Args:
            name: Str. Name of parameter. Must be unique.
            min_value: Int. Lower limit of range (included).
            max_value: Int. Upper limit of range (included).
            default: Default value to return for the parameter. If unspecified, the default value will be None.
            rank: Int, default 0, Importance ordering of parameters, smaller is more important, rank should be greater than 1.
        """
        if name not in self._space:
            self._space[name] = {'mode':'Int', 'default':default, 'rank':rank, 'values':list(range(min_value, max_value+1))}
            if rank>0:
                self._rank[rank].append(name)
            self.params[name] = round(random.uniform(min_value, max_value)) if default is None else default
    
    def _rank_list_func(self):
        rank = sorted(self._rank)
        for i in rank:
            for v in itertools.product(*[self._space[j]['values'] for j in self._rank[i]]):
                grid = {name:name_value for name, name_value in zip(self._rank[i], v)}
                for n in list(grid):
                    if n in self._dependent:
                        for m in self._dependent[n]:
                            grid.update({m:self._space[m]['function'](grid[n])})
                self._rank_list.append(grid.copy())
        self._cardinality = len(self._rank_list)
    
    def update(self, best_params):
        """params update
        
        Args:
            best_params: a best params dict.
        """
        if self._update_init:
            if not self._rank_list:
                self._rank_list_func()
            self._update_init = False
        else:
            if best_params is not None:
                self.best_params = best_params.copy()
                self.params = best_params.copy()
            self.params.update(self._rank_list[self._update_nums-1])
        self.params_history[self._update_nums] = self.params.copy()
        self._update_nums += 1
        
    def cardinality(self):
        """number of grid searches."""
        if not self._rank_list:
            self._rank_list_func()
        return self._cardinality
    
    def grid_space(self):
        """grid search space."""
        return self._rank_list
    
    def from_HyperParameters(self, hp):
        """update HyperParametersGrid class.
        
        Args:
            hp: a HyperParametersGrid class.
        """
        for name, config in hp._space.items():
            self._space[name] = config
            self.params[name] = hp.params[name]
        for rank, name_list in hp._rank.items():
            if rank not in self._rank:
                self._rank[rank].append(name_list)
                continue
            for name in name_list:
                for r, n in self._rank.items():
                    if rank==r:
                        if name not in self._rank[rank]:
                            self._rank[rank].append(name)
                    else:
                        if name in self._rank[rank]:
                            self._rank[rank].remove(name)
                        


import json
from collections import defaultdict

from linora.utils._config import Config


class Coordinate():
    def __init__(self):
        self._params = Config()
        self._params.ydata = defaultdict(defaultdict)
        self._params.annotate = dict()
        self._params.figure = {'figsize':(12.8, 7.2)}
        
        self._params.axis = {'normal':{'axis':None, 'xinvert':False, 'yinvert':False, 'xtick':{}, 'ytick':{},
                             'xlabel':None, 'ylabel':None, 'xtickposition':None, 'ytickposition':None}}
        
        self._params.label = {'normal':{'xlabel':{'xlabel':None}, 'ylabel':{'ylabel':None}}}
        
        self._params.legend = dict()
        self._params.spine = {'show':{}, 'alpha':{}, 'color':{}, 'width':{}, 'style':{}, 'position':{}}
        self._params.text = dict()
        self._params.theme = 'ggplot'
        self._params.title = {'label':None}
        self._params.twin = dict()
        
        self._params.set_label = True
        self._params.colorbar = set()
        
    def get_config(self, json_path=None):
        config = {
            'mode': 'plot',
            'annotate': self._params.annotate,
            'ydata': self._params.ydata,
            'theme': self._params.theme,
            'figure': self._params.figure,
            'axis': self._params.axis,
            'label': self._params.label,
            'legend': self._params.legend,
            'spine': self._params.spine,
            'text': self._params.text,
            'title': self._params.title,
            'twin': self._params.twin,
            'set_label': self._params.set_label,
            'colorbar': self._params.colorbar,
        }
        if json_path is not None:
            with open(json_path, 'w') as f:
                json.dump(config, f)
        return config
    
    def set_config(self, config):
        if isinstance(config, str):
            assert config.endswith('.json'), f'{config} not a json file.'
            with open(config) as f:
                config = json.load(f)
        assert isinstance(config, dict), f'{config} not a dict.'
        assert config['mode']=='plot', 'config info not match.'
        self._params.ydata = config['ydata']
        self._params.annotate = config['annotate']
        self._params.theme = config['theme']
        self._params.figure = config['figure']
        self._params.axis = config['axis']
        self._params.label = config['label']
        self._params.legend = config['legend']
        self._params.spine = config['spine']
        self._params.text = config['text']
        self._params.title = config['title']
        self._params.twin = config['twin']
        self._params.set_label = config['set_label']
        self._params.colorbar = config['colorbar']
        return self
            
    def render(self, image_path=None, if_show=True, **kwargs):
        """show and save plot."""
        fig = self._execute()
        if image_path is not None:
            fig.savefig(image_path, **kwargs)
        if if_show:
            return fig.show()
    
    def set_annotate(self, text, xy, xytext=None, xycoords=None, textcoords=None, 
                     arrowprops=None, annotation_clip=None, **kwargs):
        """Annotate the point *xy* with text *text*.
        
        Args:
            text : str, The text of the annotation.

            xy : (float, float)
                The point *(x, y)* to annotate. The coordinate system is determined
                by *xycoords*.

            xytext : (float, float), default: *xy*
                The position *(x, y)* to place the text at. The coordinate system
                is determined by *textcoords*.

            xycoords : str or `.Artist` or `.Transform` or callable or (float, float), default: 'data'

                The coordinate system that *xy* is given in. The following types
                of values are supported:

                - One of the following strings:

                  ==================== ============================================
                  Value                Description
                  ==================== ============================================
                  'figure points'      Points from the lower left of the figure
                  'figure pixels'      Pixels from the lower left of the figure
                  'figure fraction'    Fraction of figure from lower left
                  'subfigure points'   Points from the lower left of the subfigure
                  'subfigure pixels'   Pixels from the lower left of the subfigure
                  'subfigure fraction' Fraction of subfigure from lower left
                  'axes points'        Points from lower left corner of axes
                  'axes pixels'        Pixels from lower left corner of axes
                  'axes fraction'      Fraction of axes from lower left
                  'data'               Use the coordinate system of the object
                                       being annotated (default)
                  'polar'              *(theta, r)* if not native 'data'
                                       coordinates
                  ==================== ============================================

                  Note that 'subfigure pixels' and 'figure pixels' are the same
                  for the parent figure, so users who want code that is usable in
                  a subfigure can use 'subfigure pixels'.

                - An `.Artist`: *xy* is interpreted as a fraction of the artist's
                  `~matplotlib.transforms.Bbox`. E.g. *(0, 0)* would be the lower
                  left corner of the bounding box and *(0.5, 1)* would be the
                  center top of the bounding box.

                - A `.Transform` to transform *xy* to screen coordinates.

                - A function with one of the following signatures::

                    def transform(renderer) -> Bbox
                    def transform(renderer) -> Transform

                  where *renderer* is a `.RendererBase` subclass.

                  The result of the function is interpreted like the `.Artist` and
                  `.Transform` cases above.

                - A tuple *(xcoords, ycoords)* specifying separate coordinate
                  systems for *x* and *y*. *xcoords* and *ycoords* must each be
                  of one of the above described types.

                See :ref:`plotting-guide-annotation` for more details.

            textcoords : str or `.Artist` or `.Transform` or callable or (float, float), default: value of *xycoords*
                The coordinate system that *xytext* is given in.

                All *xycoords* values are valid as well as the following
                strings:

                =================   =========================================
                Value               Description
                =================   =========================================
                'offset points'     Offset (in points) from the *xy* value
                'offset pixels'     Offset (in pixels) from the *xy* value
                =================   =========================================

            arrowprops : dict, optional
                The properties used to draw a `.FancyArrowPatch` arrow between the
                positions *xy* and *xytext*. Note that the edge of the arrow
                pointing to *xytext* will be centered on the text itself and may
                not point directly to the coordinates given in *xytext*.

                If *arrowprops* does not contain the key 'arrowstyle' the
                allowed keys are:

                ==========   ======================================================
                Key          Description
                ==========   ======================================================
                width        The width of the arrow in points
                headwidth    The width of the base of the arrow head in points
                headlength   The length of the arrow head in points
                shrink       Fraction of total length to shrink from both ends
                ?            Any key to :class:`matplotlib.patches.FancyArrowPatch`
                ==========   ======================================================

                If *arrowprops* contains the key 'arrowstyle' the
                above keys are forbidden.  The allowed values of
                ``'arrowstyle'`` are:

                ============   =============================================
                Name           Attrs
                ============   =============================================
                ``'-'``        None
                ``'->'``       head_length=0.4,head_width=0.2
                ``'-['``       widthB=1.0,lengthB=0.2,angleB=None
                ``'|-|'``      widthA=1.0,widthB=1.0
                ``'-|>'``      head_length=0.4,head_width=0.2
                ``'<-'``       head_length=0.4,head_width=0.2
                ``'<->'``      head_length=0.4,head_width=0.2
                ``'<|-'``      head_length=0.4,head_width=0.2
                ``'<|-|>'``    head_length=0.4,head_width=0.2
                ``'fancy'``    head_length=0.4,head_width=0.4,tail_width=0.4
                ``'simple'``   head_length=0.5,head_width=0.5,tail_width=0.2
                ``'wedge'``    tail_width=0.3,shrink_factor=0.5
                ============   =============================================

                Valid keys for `~matplotlib.patches.FancyArrowPatch` are:

                ===============  ==================================================
                Key              Description
                ===============  ==================================================
                arrowstyle       the arrow style
                connectionstyle  the connection style
                relpos           default is (0.5, 0.5)
                patchA           default is bounding box of the text
                patchB           default is None
                shrinkA          default is 2 points
                shrinkB          default is 2 points
                mutation_scale   default is text size (in points)
                mutation_aspect  default is 1.
                ?                any key for :class:`matplotlib.patches.PathPatch`
                ===============  ==================================================

                Defaults to None, i.e. no arrow is drawn.

            annotation_clip : bool or None, default: None
                Whether to draw the annotation when the annotation point *xy* is
                outside the axes area.

                - If *True*, the annotation will only be drawn when *xy* is
                  within the axes.
                - If *False*, the annotation will always be drawn.
                - If *None*, the annotation will only be drawn when *xy* is
                  within the axes and *xycoords* is 'data'.
        """
        if xytext is not None:
            kwargs['xytext'] = xytext
        if xycoords is not None:
            kwargs['xycoords'] = xycoords
        if textcoords is not None:
            kwargs['textcoords'] = textcoords
        if arrowprops is not None:
            kwargs['arrowprops'] = arrowprops
        if annotation_clip is not None:
            kwargs['annotation_clip'] = annotation_clip
        kwargs['text'] = text
        kwargs['xy'] = xy
        self._params.annotate[len(self._params.annotate)] = kwargs
        return self
        
    def set_axis(self, axis=None, xmin=None, xmax=None, ymin=None, ymax=None,
                 xlabel=None, ylabel=None,
                 invert=False, xinvert=False, yinvert=False, 
                 tickshow=True, xtickshow=True, ytickshow=True, 
                 tickwhich=None, xtickwhich=None, ytickwhich=None, 
                 tickcolor=None, xtickcolor=None, ytickcolor=None, 
                 tickheight=None, xtickheight=None, ytickheight=None, 
                 tickwidth=None, xtickwidth=None, ytickwidth=None, 
                 tickpad=None, xtickpad=None, ytickpad=None, 
                 tickloc=None, xtickloc=None, ytickloc=None,
                 xtickposition=None, ytickposition=None,
                 labelsize=None, xlabelsize=None, ylabelsize=None, 
                 labelcolor=None, xlabelcolor=None, ylabelcolor=None, 
                 labelrotate=None, xlabelrotate=None, ylabelrotate=None, 
                 gridcolor=None, xgridcolor=None, ygridcolor=None, 
                 gridalpha=None, xgridalpha=None, ygridalpha=None, 
                 gridwidth=None, xgridwidth=None, ygridwidth=None, 
                 gridstyle=None, xgridstyle=None, ygridstyle=None, 
                 tickbottom=None, ticktop=None, tickleft=None, tickright=None,
                 labelbottom=None, labeltop=None, labelleft=None, labelright=None,
                 twin=None,
                ):
        """Convenience method to get or set some axis properties.
        
        Args:
            axis: bool or str, If a bool, turns axis lines and labels on or off. 
                  If a string, possible values are:
                    ======== ==========================================================
                    Value    Description
                    ======== ==========================================================
                    'on'     Turn on axis lines and labels. Same as ``True``.
                    'off'    Turn off axis lines and labels. Same as ``False``.
                    'equal'  Set equal scaling (i.e., make circles circular) by
                             changing axis limits. This is the same as
                             ``ax.set_aspect('equal', adjustable='datalim')``.
                             Explicit data limits may not be respected in this case.
                    'scaled' Set equal scaling (i.e., make circles circular) by
                             changing dimensions of the plot box. This is the same as
                             ``ax.set_aspect('equal', adjustable='box', anchor='C')``.
                             Additionally, further autoscaling will be disabled.
                    'tight'  Set limits just large enough to show all data, then
                             disable further autoscaling.
                    'auto'   Automatic scaling (fill plot box with data).
                    'image'  'scaled' with axis limits equal to data limits.
                    'square' Square plot; similar to 'scaled', but initially forcing
                             ``xmax-xmin == ymax-ymin``.
                    ======== ==========================================================
                    if axis is set, other parameters are invalid.
            xmin: float, The left xlim in data coordinates.
            xmax: float, The right xlim in data coordinates.
            ymin: float, The bottom ylim in data coordinates.
            ymax: float, The top ylim in data coordinates.
            xlabel: list or [list, list] of x-axis tick label.
            ylabel: list or [list, list] of y-axis tick label.
            invert: invert xy-axis.
            xinvert: invert x-axis.
            yinvert: invert y-axis.
            tickshow: show xy-axis tick.
            xtickshow: show x-axis tick.
            ytickshow: show y-axis tick.
            tickwhich: {'major', 'minor', 'both'}, default: 'major', 
                The group of xy-axis ticks to which the parameters are applied.
            xtickwhich: {'major', 'minor', 'both'}, default: 'major', 
                The group of x-axis ticks to which the parameters are applied.
            ytickwhich: {'major', 'minor', 'both'}, default: 'major', 
                The group of y-axis ticks to which the parameters are applied.
            tickcolor: xy-axis tick color.
            xtickcolor: x-axis tick color.
            ytickcolor: y-axis tick color.
            tickheight: xy-axis tick height.
            xtickheight: x-axis tick height.
            ytickheight: y-axis tick height.
            tickwidth: xy-axis tick width.
            xtickwidth: x-axis tick width.
            ytickwidth: y-axis tick width.
            tickpad: xy-axis distance in points between tick and label.
            xtickpad: x-axis distance in points between tick and label.
            ytickpad: y-axis distance in points between tick and label.
            tickloc: {'in', 'out', 'inout'}, xy-axis puts ticks position.
            xtickloc: {'in', 'out', 'inout'}, x-axis puts ticks position.
            ytickloc: {'in', 'out', 'inout'}, y-axis puts ticks position.
            xtickposition: {'top', 'bottom', 'both', 'default', 'none'}, ticks position.
            ytickposition: {'left', 'right', 'both', 'default', 'none'}, ticks position.
            labelsize: xy-axis label font size.
            xlabelsize: x-axis label font size.
            ylabelsize: y-axis label font size.
            labelcolor: xy-axis label font color.
            xlabelcolor: x-axis label font color.
            ylabelcolor: y-axis label font color.
            labelrotate: xy-axis label font rotate.
            xlabelrotate: x-axis label font rotate.
            ylabelrotate: y-axis label font rotate.
            gridcolor: xy-axis grid line color.
            xgridcolor: x-axis grid line color.
            ygridcolor: y-axis grid line color.
            gridalpha: xy-axis grid line alpha.
            xgridalpha: x-axis grid line alpha.
            ygridalpha: y-axis grid line alpha.
            gridwidth: xy-axis grid line width.
            xgridwidth: x-axis grid line width.
            ygridwidth: y-axis grid line width.
            gridstyle: xy-axis grid line style.
            xgridstyle: x-axis grid line style.
            ygridstyle: y-axis grid line style.
            tickbottom: Whether to draw bottom ticks.
            ticktop: Whether to draw top ticks.
            tickleft: Whether to draw the left ticks.
            tickright: Whether to draw the right ticks.
            labelbottom: Whether to draw bottom tick labels.
            labeltop: Whether to draw top tick labels.
            labelleft: Whether to draw left tick labels.
            labelright: Whether to draw right tick labels.
            twin: twin Axes, 'x' or 'y'.
        """
        if twin is not None:
            self._params.axis[twin] = {'axis':None, 'xinvert':False, 'yinvert':False, 'xtick':{}, 'ytick':{},
                                       'xlabel':None, 'ylabel':None, 'xtickposition':None, 'ytickposition':None}
        mode = twin if twin is not None else 'normal'
        if axis is not None:
            self._params.axis[mode]['axis'] = axis
        elif xmin is not None or xmax is not None or ymin is not None or ymax is not None:
            self._params.axis[mode]['axis'] = [xmin, xmax, ymin, ymax]
        self._params.axis[mode]['xinvert'] = invert or xinvert
        self._params.axis[mode]['yinvert'] = invert or yinvert
        if xlabel is not None:
            self._params.axis[mode]['xlabel'] = xlabel
        if ylabel is not None:
            self._params.axis[mode]['ylabel'] = ylabel
        if xtickposition is not None:
            self._params.axis[mode]['xtickposition'] = xtickposition
        if ytickposition is not None:
            self._params.axis[mode]['ytickposition'] = ytickposition
        
        self._set_axis(xtickwhich, ytickwhich, tickwhich, 'which', mode)
        self._set_axis(xtickcolor, ytickcolor, tickcolor, 'color', mode)
        self._set_axis(xtickheight, ytickheight, tickheight, 'length', mode)
        self._set_axis(xtickwidth, ytickwidth, tickwidth, 'width', mode)
        self._set_axis(xtickloc, ytickloc, tickloc, 'direction', mode)
        self._set_axis(xtickpad, ytickpad, tickpad, 'pad', mode)
        self._set_axis(xlabelsize, ylabelsize, labelsize, 'labelsize', mode)
        self._set_axis(xlabelcolor, ylabelcolor, labelcolor, 'labelcolor', mode)
        self._set_axis(xlabelrotate, ylabelrotate, labelrotate, 'labelrotation', mode)
        self._set_axis(xgridcolor, ygridcolor, gridcolor, 'grid_color', mode)
        self._set_axis(xgridalpha, ygridalpha, gridalpha, 'grid_alpha', mode)
        self._set_axis(xgridwidth, ygridwidth, gridwidth, 'grid_linewidth', mode)
        self._set_axis(xgridstyle, ygridstyle, gridstyle, 'grid_linestyle', mode)
            
        if tickbottom is not None:
            self._params.axis[mode]['xtick']['bottom'] = tickbottom
        if ticktop is not None:
            self._params.axis[mode]['xtick']['top'] = ticktop
        if tickleft is not None:
            self._params.axis[mode]['ytick']['left'] = tickleft
        if tickright is not None:
            self._params.axis[mode]['ytick']['right'] = tickright
        if labelbottom is not None:
            self._params.axis[mode]['xtick']['labelbottom'] = labelbottom
        if labeltop is not None:
            self._params.axis[mode]['xtick']['labeltop'] = labeltop
        if labelleft is not None:
            self._params.axis[mode]['ytick']['labelleft'] = labelleft
        if labelright is not None:
            self._params.axis[mode]['ytick']['labelright'] = labelright

        if not (tickshow and xtickshow):
            self._params.axis[mode]['xtick'] = {'length':0, 'labelsize':0, 'which':'both'}
        if not (tickshow and ytickshow):
            self._params.axis[mode]['ytick'] = {'length':0, 'labelsize':0, 'which':'both'}
        return self
    
    def _set_axis(self, x, y, xy, s, mode):
        if x is not None:
            self._params.axis[mode]['xtick'][s] = x['mode'] if isinstance(x, dict) else x
        if y is not None:
            self._params.axis[mode]['ytick'][s] = y['mode'] if isinstance(y, dict) else y
        if xy is not None:
            self._params.axis[mode]['xtick'][s] = xy['mode'] if isinstance(xy, dict) else xy
            self._params.axis[mode]['ytick'][s] = xy['mode'] if isinstance(xy, dict) else xy
        
    def set_figure(self, width=12.8, height=7.2, dpi=None, facecolor=None, edgecolor=None, frameon=True, clear=False):
        """Add figure config.
        
        Args:
            width: float, figure size width in inches.
            height: float, figure size height in inches.
            dpi: float, The resolution of the figure in dots-per-inch.
            facecolor: color, The background color.
            edgecolor: color, The border color.
            frameon: bool, default: True, If False, suppress drawing the figure frame.
            clear: bool, default: False, If True and the figure already exists, then it is cleared.
        """
        kwargs = {'figsize':(width, height), 'dpi':dpi, 'facecolor':facecolor, 
                  'edgecolor':edgecolor, 'frameon':frameon, 'clear':clear}
        self._params.figure.update(kwargs)
        return self
    
    def set_label(self, xlabel=None, ylabel=None, 
                  loc=None, xloc=None, yloc=None, 
                  pad=None, xpad=None, ypad=None,
                  fontsize=None, xfontsize=None, yfontsize=None, 
                  fontcolor=None, xfontcolor=None, yfontcolor=None, 
                  fontfamily=None, xfontfamily=None, yfontfamily=None, 
                  fontstyle=None, xfontstyle=None, yfontstyle=None, 
                  twin=None,
                 ):
        """Set the label for the x-axis and y-axis.
        
        Args:
            xlabel: str, The label text.
            ylabel: str, The label text.
            loc: {'left', 'center', 'right'}, The xy label position.
            xloc: {'left', 'center', 'right'}, The xlabel position.
            yloc: {'bottom', 'center', 'top'}, The ylabel position.
            pad: float, Spacing in points from the axes bounding box including ticks and tick xy labels.
            xpad : float, Spacing in points from the axes bounding box including ticks and tick x labels.
            ypad : float, Spacing in points from the axes bounding box including ticks and tick y labels.
            fontsize: xy label font size.
            xfontsize: x label font size.
            yfontsize: y label font size.
            fontcolor: xy label font color.
            xfontcolor: x label font color.
            yfontcolor: y label font color.
            fontfamily:
            xfontfamily:
            yfontfamily:
            fontstyle: xy label font style.
            xfontstyle: x label font style.
            yfontstyle: y label font style.
            twin: twin Axes, 'x' or 'y'.
        """
        if twin is not None:
            self._params.label[twin] = {'xlabel':{'xlabel':None}, 'ylabel':{'ylabel':None}}
        mode = twin if twin is not None else 'normal'
        self._set_label(xlabel, y=None, xy=None, xkey='xlabel', mode=mode)
        self._set_label(x=None, y=ylabel, xy=None, ykey='ylabel', mode=mode)
        self._set_label(x=xloc, y=yloc, xy=loc, xkey='loc', ykey='loc', mode=mode)
        self._set_label(x=xpad, y=ypad, xy=pad, xkey='labelpad', ykey='labelpad', mode=mode)
        self._set_label(x=xfontsize, y=yfontsize, xy=fontsize, xkey='fontsize', ykey='fontsize', mode=mode)
        self._set_label(x=xfontcolor, y=yfontcolor, xy=fontcolor, xkey='color', ykey='color', mode=mode)
        self._set_label(x=xfontfamily, y=yfontfamily, xy=fontfamily, xkey='fontfamily', ykey='fontfamily', mode=mode)
        self._set_label(x=xfontstyle, y=yfontstyle, xy=fontstyle, xkey='fontstyle', ykey='fontstyle', mode=mode)
        return self
    
    def _set_label(self, x, y, xy, xkey=None, ykey=None, mode='normal'):
        if x is not None:
            self._params.label[mode]['xlabel'][xkey] = x['mode'] if isinstance(x, dict) else x
        if y is not None:
            self._params.label[mode]['ylabel'][ykey] = y['mode'] if isinstance(y, dict) else y
        if xy is not None:
            self._params.label[mode]['xlabel'][xkey] = xy['mode'] if isinstance(xy, dict) else xy
            self._params.label[mode]['ylabel'][ykey] = xy['mode'] if isinstance(xy, dict) else xy
    
    def set_legend(self, loc='best', **kwargs):
        """Place a legend on the Axes.
        
        Args:
            loc: str or pair of floats, The location of the legend.
                The strings ``'upper left', 'upper right', 'lower left', 'lower right'``
                place the legend at the corresponding corner of the axes/figure.

                The strings ``'upper center', 'lower center', 'center left', 'center right'``
                place the legend at the center of the corresponding edge of the axes/figure.
                The string ``'center'`` places the legend at the center of the axes/figure.

                The string ``'best'`` places the legend at the location, among the nine
                locations defined so far, with the minimum overlap with other drawn artists.  
                This option can be quite slow for plots with large amounts of data; 
                your plotting speed may benefit from providing a specific location.

                The location can also be a 2-tuple giving the coordinates of the lower-left
                corner of the legend in axes coordinates (in which case *bbox_to_anchor*
                will be ignored).
            fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
                      The font size of the legend. If the value is numeric the size will be the absolute 
                      font size in points. String values are relative to the current default font size.
            

            numpoints: int, The number of marker points in the legend when creating a legend.
            markerscale: float, The relative size of legend markers compared with the originally drawn ones.
            markerfirst: bool, default: True
                        If *True*, legend marker is placed to the left of the legend label.
                        If *False*, legend marker is placed to the right of the legend label.

            frameon : bool, Whether the legend should be drawn on a patch (frame).

            fancybox : bool, Whether round edges should be enabled around the `~.FancyBboxPatch` which
                makes up the legend's background.

            shadow : bool, Whether to draw a shadow behind the legend.

            framealpha : float, The alpha transparency of the legend's background.
                         If *shadow* is activated and *framealpha* is ``None``, the default value is ignored.

            facecolor : "inherit" or color, default: :rc:`legend.facecolor`
                The legend's background color.
                If ``"inherit"``, use :rc:`axes.facecolor`.

            edgecolor : "inherit" or color, default: :rc:`legend.edgecolor`
                The legend's background patch edge color.
                If ``"inherit"``, use take :rc:`axes.edgecolor`.

            mode : {"expand", None}
                If *mode* is set to ``"expand"`` the legend will be horizontally
                expanded to fill the axes area (or *bbox_to_anchor* if defines
                the legend's size).

            bbox_transform : None or `matplotlib.transforms.Transform`
                The transform for the bounding box (*bbox_to_anchor*). For a value
                of ``None`` (default) the Axes'
                :data:`~matplotlib.axes.Axes.transAxes` transform will be used.

            title : str or None, The legend's title. Default is no title (``None``).
            title_fontsize: int or {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
                            The font size of the legend's title.
            borderpad : float, The fractional whitespace inside the legend border, in font-size units.
            labelspacing : float, The vertical space between the legend entries, in font-size units.
            handlelength : float, The length of the legend handles, in font-size units.
            handletextpad : float, The pad between the legend handle and text, in font-size units.
            borderaxespad : float, The pad between the axes and legend border, in font-size units.
            columnspacing : float, The spacing between columns, in font-size units.

            handler_map : dict or None
                The custom dictionary mapping instances or types to a legend
                handler. This *handler_map* updates the default handler map
                found at `matplotlib.legend.Legend.get_legend_handler_map`.
        """
#         'legend.borderaxespad': 0.5,
# 'legend.borderpad': 0.4,
# 'legend.columnspacing': 2.0,
# 'legend.edgecolor': '0.8',
# 'legend.facecolor': 'inherit',
# 'legend.fancybox': True,
# 'legend.fontsize': 'medium',
# 'legend.framealpha': 0.8,
# 'legend.frameon': True,
# 'legend.handleheight': 0.7,
# 'legend.handlelength': 2.0,
# 'legend.handletextpad': 0.8,
# 'legend.labelspacing': 0.5,
# 'legend.loc': 'best',
# 'legend.markerscale': 1.0,
# 'legend.numpoints': 1,
# 'legend.scatterpoints': 1,
# 'legend.shadow': False,
# 'legend.title_fontsize': None,
        kwargs['loc'] = loc
        self._params.legend.update(kwargs)
        return self
        
    def set_spine(self, spineshow=None, leftshow=None, rightshow=None, topshow=None, bottomshow=None,
                  spinecolor=None, leftcolor=None, rightcolor=None, topcolor=None, bottomcolor=None,
                  spinewidth=None, leftwidth=None, rightwidth=None, topwidth=None, bottomwidth=None,
                  spinestyle=None, leftstyle=None, rightstyle=None, topstyle=None, bottomstyle=None,
                  spinealpha=None, leftalpha=None, rightalpha=None, topalpha=None, bottomalpha=None,
                  leftloc=None, rightloc=None, toploc=None, bottomloc=None,
                 ):
        """Set a wireframes for the Axes.
        
        Args:
            spineshow: bool, whether to show all wireframes.
            leftshow: bool, whether to show left wireframes.
            rightshow: bool, whether to show right wireframes.
            topshow: bool, whether to show top wireframes.
            bottomshow: bool, whether to show bottom wireframes.
            spinecolor: all wireframes color.
            leftcolor: left wireframes color.
            rightcolor: right wireframes color.
            topcolor: top wireframes color.
            bottomcolor: bottom wireframes color.
            spinewidth: all wireframes width.
            leftwidth: left wireframes width.
            rightwidth: right wireframes width.
            topwidth: top wireframes width.
            bottomwidth: bottom wireframes width.
            spinestyle: all wireframes style.
            leftstyle: left wireframes style.
            rightstyle: right wireframes style.
            topstyle: top wireframes style.
            bottomstyle: bottom wireframes style.
            spinealpha: all wireframes alpha.
            leftalpha: left wireframes alpha.
            rightalpha: right wireframes alpha.
            topalpha: top wireframes alpha.
            bottomalpha: bottom wireframes alpha.
            leftloc: 2 tuple of (position type, amount), left wireframes position.
                The position types are :
                * 'outward': place the spine out from the data area by the specified number of points.
                * 'axes': place the spine at the specified Axes coordinate (0 to 1).
                * 'data': place the spine at the specified data coordinate.
            rightloc: 2 tuple of (position type, amount), right wireframes position.
            toploc: 2 tuple of (position type, amount), top wireframes position.
            bottomloc: 2 tuple of (position type, amount), bottom wireframes position.
        """
        self._set_spine(spineshow, leftshow, rightshow, topshow, bottomshow, 'show')
        self._set_spine(spinecolor, leftcolor, rightcolor, topcolor, bottomcolor, 'color')
        self._set_spine(spinewidth, leftwidth, rightwidth, topwidth, bottomwidth, 'width')
        self._set_spine(spinestyle, leftstyle, rightstyle, topstyle, bottomstyle, 'style')
        self._set_spine(spinealpha, leftalpha, rightalpha, topalpha, bottomalpha, 'alpha')
        self._set_spine(None, leftloc, rightloc, toploc, bottomloc, 'position')
        return self
    
    def _set_spine(self, spine, left, right, top, bottom, key):
        if left is not None:
            self._params.spine[key]['left'] = left['mode'] if isinstance(left, dict) else left
        if right is not None:
            self._params.spine[key]['right'] = right['mode'] if isinstance(right, dict) else right
        if top is not None:
            self._params.spine[key]['top'] = top['mode'] if isinstance(top, dict) else top
        if bottom is not None:
            self._params.spine[key]['bottom'] = bottom['mode'] if isinstance(bottom, dict) else bottom
        if spine is not None:
            self._params.spine[key]['left'] = spine['mode'] if isinstance(spine, dict) else spine
            self._params.spine[key]['right'] = spine['mode'] if isinstance(spine, dict) else spine
            self._params.spine[key]['top'] = spine['mode'] if isinstance(spine, dict) else spine
            self._params.spine[key]['bottom'] = spine['mode'] if isinstance(spine, dict) else spine
        
    def set_text(self, x, y, text, **kwargs):
        """Add text to the Axes.
        
        Args:
            x, y : float, The position to place the text. 
            text : str, The text.

        """
        kwargs['x'] = x
        kwargs['y'] = y
        kwargs['s'] = text
        self._params.text[len(self._params.text)] = kwargs
        return self
        
    def set_theme(self, theme):
        """Set a theme for the Axes.
        
        Args:
            theme: str, figure theme.
        """
        self._params.theme = theme
        return self
    
    def set_title(self, title, loc='center', fontsize='large', fontcolor='#000000', 
                  titlepad=6., titley=None, fontweight='normal'):
        """Set a title for the Axes.
        
        Args:
            title: str, Text to use for the title.
            loc : {'center', 'left', 'right'}, Which title to set.
            titlepad : float, The offset of the title from the top of the Axes, in points.
            titley : float, Vertical Axes loation for the title (1.0 is the top).
                     If None (the default), titley is determined automatically to avoid decorators on the Axes.
        """
        if isinstance(fontcolor, dict):
            fontcolor = fontcolor['mode']
        kwargs = {'label':title, 'loc':loc, 'pad':titlepad, 'y':titley,
                  'fontdict':{'fontsize':fontsize, 'fontweight':fontweight, 'color':fontcolor,
                              'verticalalignment': 'baseline', 'horizontalalignment': loc}}
        self._params.title = kwargs
        return self
        
    def set_twin(self, name, axis='x'):
        """Create a twin Axes sharing the x-y axis.
        
        Args:
            name: data name.
            axis: 'x' or 'y'.
        """
        self._params.twin[name] = axis
        return self
    



import numpy as np

from linora.utils._config import Config

__all__ = ['TerminateOnNaN', 'TerminateOverfitted']


class TerminateOnNaN():
    """Callback that terminates training when a NaN loss is encountered.
    
    Args:
        monitor: Quantity to be monitored.
    """
    def __init__(self, monitor):
        self._params = Config()
        self._params.monitor = monitor
        self._params.state = False
        self._params.name = 'TerminateOnNaN'
        
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.monitor in log:
            if np.isnan(log[self._params.monitor]).any():
                self._params.state = True
                

class TerminateOverfitted():
    '''Terminates training when the running loss is smaller than the theoretical best, 
    which is strong indication that the model will end up overfitting.
    
    Args:
        monitor: Quantity to be monitored.
        theory_best : float, The theoretical-smallest loss achievable without overfiting.
        mode: One of {"min", "max"}. In min mode, training will stop when the quantity monitored has stopped decreasing; 
            in "max" mode it will stop when the quantity monitored has stopped increasing.
    '''
    def __init__(self, monitor, theory_best, mode='min'):
        self._params = Config()
        self._params.monitor = monitor
        self._params.theory_best = theory_best
        self._params.state = False
        self._params.name = 'TerminateOverfitted'

    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.monitor in log:
            if self._params.mode=='min':
                if log[self._params.monitor]<self._params.theory_best:
                    self._params.state = True
            else:
                if log[self._params.monitor]>self._params.theory_best:
                    self._params.state = True
        



import sys
import time
import logging
import itertools
from collections import Counter
from multiprocessing import cpu_count

import numpy as np
import xgboost as xgb
from sklearn.model_selection import StratifiedKFold, train_test_split


def GridSearch(feature, label, group, metrics, scoring=0.5, cv=5, cv_num=3,
               metrics_min=True, speedy=True, speedy_param=(20000, 0.3), gpu=False):
    """XGBRanker model params search use GridSearch method.
    
    Args:
        feature: pandas dataframe, model's feature.
        label: pandas series, model's label.
        loss: XGBRanker param 'objective'.
        metrics: model metrics function.
        scoring: metrics error opt base line value.
        cv: cross validation fold.
        cv_num: minimum cross validation fold.
        metrics_min: metrics value whether the smaller the better.
        speedy: whether use speedy method.
        speedy_param: if use speedy method, test_size will be set, 
                      test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
        gpu: whether use gpu.
    Returns:
        a best XGBRanker model params dict.
    Raises:
        params error.
    """
    def product(x):
        if len(x)==1:
            return itertools.product(x[0])
        elif len(x)==2:
            return itertools.product(x[0], x[1])
        else:
            return itertools.product(x[0], x[1], x[2])
    start = time.time()
    if gpu:
        raise "XGBRanker is not supported currently."
    if speedy:
        test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
    tree_method = 'gpu_hist' if gpu else 'auto'
    n_job = 1 if gpu else int(np.ceil(cpu_count()*0.9))
    weight_dict = Counter(label)
    if len(weight_dict)==2:
        weight = int(np.ceil(weight_dict[min(weight_dict)]/weight_dict[max(weight_dict)]))
    else:
        weight_dict = {j:i for i,j in weight_dict.items()}
        weight = int(np.ceil(weight_dict[max(weight_dict)]/weight_dict[min(weight_dict)]))
    params = {'learning_rate': 0.1, 'n_estimators': 300, 'max_depth': 5, 'min_child_weight': 1,
              'reg_alpha': 0, 'reg_lambda': 1, 'gamma': 0,
              'subsample': 0.8, 'colsample_bytree': 0.8, 'colsample_bylevel': 0.8,
              'max_delta_step': 0, 'scale_pos_weight': weight,
              'n_jobs':n_job, 'random_state': 27, 'objective': 'rank:pairwise', 'tree_method':tree_method}
    cv_params = {'param1':{'n_estimators': list(range(100, 850, 50))},
                 'param2':{'max_depth': [3, 4, 5, 6, 7],
                           'min_child_weight': [1, 2, 3, 4, 5]},
                 'param3':{'gamma': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]},
                 'param4':{'subsample': [0.6, 0.7, 0.8, 0.9],
                           'colsample_bytree': [0.6, 0.7, 0.8, 0.9],
                           'colsample_bylevel': [0.6, 0.7, 0.8, 0.9]},
                 'param5':{'reg_alpha': [0.05, 0.1, 1, 2, 3],
                           'reg_lambda': [0.05, 0.1, 1, 2, 3]},
                 'param6':{'max_delta_step': list(np.linspace(0, 10, 11, dtype='int')),
                           'scale_pos_weight': list(np.linspace(1, weight, weight, dtype='int'))},
                 'param7':{'learning_rate': [0.01, 0.03, 0.05, 0.07, 0.1, 0.2]}}
    for _, cv_param in cv_params.items():
        cv_param_name = [i for i in cv_param]
        cv_param_value = [cv_param[i] for i in cv_param_name]
        cv_param_iter = product(cv_param_value)
        for value in cv_param_iter:
            params.update({name:name_value for name, name_value in zip(cv_param_name, value)})
            model = xgb.XGBRanker(**params)
            score = []
            if speedy:
                for i in range(cv_num):
                    X_train, X_test, y_train, y_test, g_train, g_test = train_test_split(feature, label, group, 
                                                                                         test_size=test_size, stratify=label, 
                                                                                         random_state=np.random.choice(range(100), 1)[0])
                    model.fit(X_train, y_train, g_train)
                    cv_pred = model.predict(X_test)
                    score.append(metrics(y_test.values, cv_pred))
            else:
                skf = StratifiedKFold(n_splits=cv, shuffle=True, random_state=np.random.choice(range(100), 1)[0])
                for n, (train_index, test_index) in enumerate(skf.split(feature, label)):
                    if n == cv_num:
                        break
                    model.fit(feature.loc[train_index], label[train_index], group[train_index])
                    cv_pred = model.predict(feature.loc[test_index])
                    score.append(metrics(label[test_index].values, cv_pred))
            cv_score = round(np.mean(score), 4)
            if metrics_min:
                if cv_score<scoring:
                    scoring = cv_score
                    best_params = params.copy()
            else:
                if cv_score>scoring:
                    scoring = cv_score
                    best_params = params.copy()
        params = best_params.copy()
        sys.stdout.write("XGBRanker grid search run time {} min, best score: {}, best param：{}\r".format(
            divmod((time.time()-start),60)[0], scoring, best_params))
        sys.stdout.flush()
    print("XGBRanker param finetuning with grid search run time: %d min %.2f s" % divmod((time.time() - start), 60))
    return best_params



import numpy as np

__all__ = ['missing_columns', 'single_columns', 'correlation_columns',
           'cv_columns']


def missing_columns(df, missing_rate=0.):
    """Find missing features
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        threshold: float, default=0.,[0,1)
            Count all features with a missing rate greater than `missing_rate`.
    
    Return:
        t : All features with a missing rate greater than `missing_rate`
    """
    assert 1>=missing_rate>=0, "`missing_rate` should be one of [0, 1]."
    t = (1-df.count()/len(df)).reset_index()
    t.columns = ['feature_name', 'missing_rate']
    return t[t.missing_rate>=missing_rate].reset_index(drop=True)


def single_columns(df):
    """Find single value features
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
    Return:
        All features with single value
    """
    return df.nunique().rename('single').reset_index().query('single==1')['index'].to_list()


def correlation_columns(df, corr_rate=0.9):
    """Find features whose correlation coefficient is greater than a certain threshold.
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        corr_rate : float, default=0.9, [0,1]
            Count all features with corr greater than `corr_rate`.
    Return:
        All features with corr greater than `corr_rate`.
    """
    t = df.corr()
    t = t.where(np.triu(np.ones(t.shape), k=1).astype(bool)).fillna(0)
    return [[i,j,t.loc[i, j]] for i in t for j in t[i].index if abs(t.loc[i, j])>corr_rate]


def cv_columns(df, cv_rate=0.):
    """Find features with coefficients of variation less than a certain threshold.
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        cv_rate: float, default=0.
            Count all features with cv less than `cv_rate`.
    
    Return:
        All features with cv less than `cv_rate`.
    """
    t = df.replace({np.PINF:np.nan, np.NINF:np.nan})
    t = ((t-t.min())/(t.max()-t.min())*900)+100
    t = (t.std()/t.mean()).reset_index()
    t.columns = ['feature_name', 'cv_rate']
    t = t[t.cv_rate<cv_rate].feature_name.tolist()
    return t




import numpy as np

from linora.metrics._classification import confusion_matrix

__all__ = ['iou_detection', 'psnr', 'ssim', 'total_variation']


def iou_detection(box1, box2, mode=None):
    """Computational object detection iou.
    
    Args:
        box1: list, [xmin1, ymin1, xmax1, ymax1]
        box2: list, [xmin2, ymin2, xmax2, ymax2]
        mode: str, one of ['C', 'D', 'E', 'G'], mean CIoU, DIoU, EIoU, GIoU, if None, is IoU.
    return:
        object detection iou values.
    """
    if isinstance(box1[0], (list,tuple)):
        box1 = [box1[0][0], box1[0][1], box1[1][0], box1[1][1]]
    if isinstance(box2[0], (list,tuple)):
        box2 = [box2[0][0], box2[0][1], box2[1][0], box2[1][1]]
    xmin1, ymin1, xmax1, ymax1 = box1
    xmin2, ymin2, xmax2, ymax2 = box2
    
    s1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    s2 = (xmax2 - xmin2) * (ymax2 - ymin2)
 
    w = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    h = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    
    iou = w * h / (s1 + s2 - w*h)
    if mode is None:
        return iou
    if mode=='G':
        area_c = (max(xmax1, xmax2) - min(xmin1, xmin2)) * (max(ymax1, ymax2) - min(ymin1, ymin2))
        return  iou - (area_c - (s1 + s2 - w*h))/area_c
    
    center1 = ((xmin1 + xmax1) / 2, (ymin1 + ymax1) / 2)
    center2 = ((xmin2 + xmax2) / 2, (ymin2 + ymax2) / 2)
    d_center = (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2

    corner1 = (min(xmin1, xmax1, xmin2, xmax2), min(ymin1, ymax1, ymin2, ymax2))
    corner2 = (max(xmin1, xmax1, xmin2, xmax2), max(ymin1, ymax1, ymin2, ymax2))
    d_corner = (corner1[0] - corner2[0]) ** 2 + (corner1[1] + corner2[1]) ** 2
    if mode=='D':
        return  iou - d_center / d_corner
    
    w1, h1 = xmax1 - xmin1, ymax1 - ymin1
    w2, h2 = xmax2 - xmin2, ymax2 - ymin2
    if mode=='C':
        v = 4 * (np.arctan(w1 / h1) - np.arctan(w2 / h2)) ** 2 / (np.pi ** 2)
        alpha = v / (1 - iou + v)
        return iou - d_center / d_corner - alpha * v
    
    if mode=='E':
        w_center = (w1-w2)**2
        w_corner = (corner1[0] - corner2[0]) ** 2

        h_center = (h1-h2)**2
        h_corner = (corner1[1] + corner2[1]) ** 2
        return iou - d_center/d_corner - w_center/w_corner - h_center/h_corner
    raise ValueError("mode must be one of [None, 'C', 'D', 'E', 'G']")


def iou_segmentation(y_true, y_pred):
    """Calculate object segmentation iou.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels.
    Returns:
        a dict, object segmentation iou values.
    """
    cm = confusion_matrix(np.array(y_true).flatten(), np.array(y_pred).flatten())
    intersection = np.diag(cm)
    union = np.sum(cm, axis=1) + np.sum(cm, axis=0) - np.diag(cm)
    IoU = intersection / union 
    MIoU = round(np.mean(IoU), 4)
    return {'IoU_mean':MIoU, 'IoU_class':IoU.round(4).to_dict()}


def psnr(image1, image2):
    """Peak Signal to Noise Ratio, PSNR.
    
    Args:
        image1: image array.
        image2: image array with same shape as image1.
    Returns:
        PSNR values.
    """
    mse = np.mean(np.square(image1.astype(np.float64) - image2.astype(np.float64)))
    return 10. * np.log10(np.square(255.) / mse)


def ssim(image1, image2):
    """structural similarity index Measure, SSIM.
    
    Args:
        image1: image array, grayscale image.
        image2: image array with same shape as image1.
    Returns:
        SSIM values.
    """
    assert len(image1.shape) == 2 and len(image2.shape) == 2 and image1.shape == image2.shape
    mu1 = image1.mean()
    mu2 = image2.mean()
    sigma1 = np.sqrt(((image1 - mu1) ** 2).mean())
    sigma2 = np.sqrt(((image2 - mu2) ** 2).mean())
    sigma12 = ((image1 - mu1) * (image2 - mu2)).mean()
    k1, k2, L = 0.01, 0.03, 255
    C1 = (k1*L) ** 2
    C2 = (k2*L) ** 2
    C3 = C2/2
    l12 = (2*mu1*mu2 + C1)/(mu1 ** 2 + mu2 ** 2 + C1)
    c12 = (2*sigma1*sigma2 + C2)/(sigma1 ** 2 + sigma2 ** 2 + C2)
    s12 = (sigma12 + C3)/(sigma1*sigma2 + C3)
    ssim = l12 * c12 * s12
    return ssim


def total_variation(image):
    """Calculate and return the total variation for image.
    
    The total variation is the sum of the absolute differences for neighboring
    pixel-values in the input images. This measures how much noise is in the image.
    Args:
        image: a NumPy array of image.
    Returns:
        a float.
    """
    if image.ndim==3:
        pixel_dif1 = image[1:, :, :] - image[:-1, :, :]
        pixel_dif2 = image[:, 1:, :] - image[:, :-1, :]
        return np.abs(pixel_dif1).sum()+np.abs(pixel_dif2).sum()
    elif image.ndim==4:
        pixel_dif1 = image[:, 1:, :, :] - image[:, :-1, :, :]
        pixel_dif2 = image[:, :, 1:, :] - image[:, :, :-1, :]
        return np.abs(pixel_dif1).sum()+np.abs(pixel_dif2).sum()
    else:
        raise ValueError('`image` must be either 3 or 4-dimensional.')


import os
import json
from multiprocessing import cpu_count

import numpy as np
import pandas as pd

from linora.utils._logger import Logger
from linora.metrics._classification import auc_roc
from linora.sample._fold import kfold, train_test_split
from linora.param_search._HyperParameters import HyperParametersRandom
from linora.param_search._config import __lightgbm_version__


class RandomSearch():
    def __init__(self):
        hp = HyperParametersRandom()
        hp.Float('learning_rate', 0.01, 0.1, default=0.1)
        hp.Int('n_estimators', 100, 850, default=200)
        hp.Choice('max_depth', [3, 4, 5, 6, 7], default=4)
        hp.Dependence('num_leaves', 'max_depth', lambda x:int(np.random.choice(range(int(2**x*0.4), int(2**x*0.6)))), default=15)
        hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
        hp.Choice('subsample', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('colsample_bytree', [0.7, 0.8, 0.9, 1. ])
        hp.Choice('random_state', [27])
        hp.Choice('boosting_type', ['gbdt'])
        hp.Choice('verbosity', [0])
        hp.Choice('n_jobs', [int(np.ceil(cpu_count()*0.8))])
        self.HyperParameter = hp
        self.best_params = dict()
        self.best_params_history = dict()

    def search(self, feature, label, vaild_data=None, sample_weight=None, 
               metrics=auc_roc, loss='regression', 
               iter_num=100, scoring=0.5, cv=5, cv_num=3, metrics_min=True, 
               speedy=True, speedy_param=(20000, 0.3), 
               save_model_dir=None, save_model_name='lgb'):
        """LGBMRegressor model params search use RandomSearch method.

        Args:
            feature: pandas dataframe, model's feature.
            label: pandas series, model's label.
            vaild_data: A list of (X, y) tuple pairs to use as validation sets, for which metrics will be computed. 
            sample_weight: pd.Series or np.array, sample weight, shape is (n,).
            metrics: model metrics function, default is `la.metircs.auc_roc`.
            loss: LGBMRegressor param 'objective'.
            scoring: metrics error opt base line value.
            cv: cross validation fold.
            cv_num: if use speedy method, minimum cross validation fold.
            metrics_min: metrics value whether the smaller the better.
            speedy: whether use speedy method.
            speedy_param: if use speedy method, test_size will be set, 
                          test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
            save_model_dir: str, save model folder.
            save_model_name: str, save model name prefix, "`lgb`_model.json" and "`lgb`_params.json".
        Returns:
            a best LGBMRegressor model params dict.
        Raises:
            params error.
        """
        logger = Logger(name='lgb')
        logger.info(f"api is deprecated and will be removed in 1.5.0")
        logger.info(f"please use la.param_search.RandomSearch")
        import warnings
        warnings.filterwarnings("ignore")
        import lightgbm as lgb
        assert lgb.__version__>=__lightgbm_version__, f'lightgbm version should be >={__lightgbm_version__}.'
        if speedy:
            test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
        self.HyperParameter.Choice('objective', [loss])

        if vaild_data is not None:
            cv_score_list = []
            
        logger.info(f"Start LGBMRegressor hyperparameter random search.")
        for i in range(1, iter_num+1):
            self.HyperParameter.update()
            model = lgb.LGBMRegressor(**self.HyperParameter.params)
            score = []
            if speedy:
                for _ in range(cv_num):
                    index_list = train_test_split(feature, label, test_size=test_size, shuffle=True, seed=np.random.choice(range(100), 1)[0])
                    weight = None if sample_weight is None else sample_weight[index_list[0]]
                    model.fit(feature.loc[index_list[0]], label[index_list[0]], sample_weight=weight)
                    cv_pred = pd.Series(model.predict(feature.loc[index_list[1]]), index=label[index_list[1]].index)
                    score.append(metrics(label[index_list[1]], cv_pred))
            else:
                index_list = kfold(feature, label, n_splits=cv, shuffle=True, seed=np.random.choice(range(100), 1)[0])
                for n, index in enumerate(index_list):
                    weight = None if sample_weight is None else sample_weight[index[0]]
                    model.fit(feature.loc[index[0]], label[index[0]], sample_weight=weight)
                    cv_pred = pd.Series(model.predict(feature.loc[index[1]]), index=label[index[1]].index)
                    score.append(metrics(label[index[1]], cv_pred))
            cv_score = np.mean(score)
            if vaild_data is not None:
                cv_score_list.append(cv_score)
                if metrics_min:
                    cv_score_list.sort()
                    if cv_score_list[int(len(cv_score_list)*0.2)]>=cv_score:
                        cv_pred = pd.Series(model.predict(vaild_data[0]), index=vaild_data[1].index)
                        cv_score = metrics(vaild_data[1], cv_pred)
                    else:
                        logger.info(f"Random search progress: {i/iter_num*100:.1f}%, best score: {scoring:.4f}", enter=False if i<iter_num else True)
                        continue
                else:
                    cv_score_list.sort(reverse=1)
                    if cv_score_list[int(len(cv_score_list)*0.2)]<=cv_score:
                        cv_pred = pd.Series(model.predict(vaild_data[0]), index=vaild_data[1].index)
                        cv_score = metrics(vaild_data[1], cv_pred)
                    else:
                        logger.info(f"Random search progress: {i/iter_num*100:.1f}%, best score: {scoring:.4f}", enter=False if i<iter_num else True)
                        continue
            if metrics_min:
                if cv_score<scoring:
                    scoring = cv_score
                    self.best_params = self.HyperParameter.params.copy()
                    self.best_params_history[i] = {'score':scoring, 'best_params':self.best_params.copy()}
                    if save_model_dir is not None:
                        model.save_model(os.path.join(save_model_dir, f"{save_model_name}_model.json"))
                        with open(os.path.join(save_model_dir, f"{save_model_name}_params.json"),'w') as f:
                            json.dump(best_params, f)
            else:
                if cv_score>scoring:
                    scoring = cv_score
                    self.best_params = self.HyperParameter.params.copy()
                    self.best_params_history[i] = {'score':scoring, 'best_params':self.best_params.copy()}
                    if save_model_dir is not None:
                        model.save_model(os.path.join(save_model_dir, f"{save_model_name}_model.json"))
                        with open(os.path.join(save_model_dir, f"{save_model_name}_params.json"),'w') as f:
                            json.dump(best_params, f)
            logger.info(f"Random search progress: {i/iter_num*100:.1f}%, best score: {scoring:.4f}", enter=False if i<iter_num else True)
        logger.info(f"LGBMRegressor random search best score: {scoring:.4f}", close=True, time_mode=1)
        return self.best_params


import numpy as np


class Line():
    def add_line(self, name, xdata, ydata, **kwargs):
        """Plot y versus x as lines and/or markers.
        
        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
            linestyle: line style, {'-', '--', '-.', ':'}.
                '-' or 'solid': solid line
                '--' or 'dashed': dashed line
                '-.' or 'dashdot': dash-dotted line
                ':' or 'dotted': dotted line
                'none', 'None', ' ', or '': draw nothing
            linecolor: line color, eg. 'blue' or '0.75' or 'g' or '#FFDD44' or (1.0,0.2,0.3) or 'chartreuse'.
            linewidth: line width.
            linelink: Set the drawstyle of the plot. The drawstyle determines how the points are connected.
                {'default', 'steps', 'steps-pre', 'steps-mid', 'steps-post'}.
                'default': the points are connected with straight lines.
                'steps-pre': The step is at the beginning of the line segment.
                'steps-mid': The step is halfway between the points.
                'steps-post': The step is at the end of the line segment.
                'steps': is equal to 'steps-pre' and is maintained for backward-compatibility.
            markfill: {'full', 'left', 'right', 'bottom', 'top', 'none'}
                'full': Fill the whole marker with the markerfacecolor.
                'left', 'right', 'bottom', 'top': Fill the marker half at the given side with the markerfacecolor. 
                                                  The other half of the marker is filled with markerfacecoloralt.
                'none': No filling.
            mark: marker style string, 
                {'.': 'point', ',': 'pixel', 'o': 'circle', 'v': 'triangle_down', '^': 'triangle_up', 
                '<': 'triangle_left', '>': 'triangle_right', '1': 'tri_down', '2': 'tri_up', '3': 'tri_left', 
                '4': 'tri_right', '8': 'octagon', 's': 'square', 'p': 'pentagon', '*': 'star', 'h': 'hexagon1', 
                'H': 'hexagon2', '+': 'plus', 'x': 'x', 'D': 'diamond', 'd': 'thin_diamond', '|': 'vline', 
                '_': 'hline', 'P': 'plus_filled', 'X': 'x_filled', 0: 'tickleft', 1: 'tickright', 2: 'tickup', 
                3: 'tickdown', 4: 'caretleft', 5: 'caretright', 6: 'caretup', 7: 'caretdown', 8: 'caretleftbase', 
                9: 'caretrightbase', 10: 'caretupbase', 11: 'caretdownbase', 
                'None': 'nothing', None: 'nothing', ' ': 'nothing', '': 'nothing'}
            markedgecolor: marker edge color.
            markedgewidth: float, marker edge width.
            markfacecolor: marker face color.
            markfacecoloralt: marker face coloralt.
            marksize: float, marker size.
            markevery: None or int or (int, int) or slice or list[int] or float or (float, float) or list[bool]
            antialiased: Set whether to use antialiased rendering.
        """
        kwargs['label'] = name
        if 'linecolor' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]#tuple([round(np.random.uniform(0, 1),1) for _ in range(3)])
        elif isinstance(kwargs['linecolor'], dict):
            kwargs['color'] = kwargs.pop('linecolor')['mode']
        else:
            kwargs['color'] = kwargs.pop('linecolor')
        if 'linelink' in kwargs:
            kwargs['drawstyle'] = kwargs.pop('linelink')
        if 'markfill' in kwargs:
            kwargs['fillstyle'] = kwargs.pop('markfill')
        if 'mark' in kwargs:
            kwargs['marker'] = kwargs.pop('mark')
        if 'markedgecolor' in kwargs:
            kwargs['markeredgecolor'] = kwargs.pop('markedgecolor')
        if 'markedgewidth' in kwargs:
            kwargs['markeredgewidth'] = kwargs.pop('markedgewidth')
        if 'markfacecolor' in kwargs:
            kwargs['markerfacecolor'] = kwargs.pop('markfacecolor')
        if 'markfacecoloralt' in kwargs:
            kwargs['markerfacecoloralt'] = kwargs.pop('markfacecoloralt')
        if 'marksize' in kwargs:
            kwargs['markersize'] = kwargs.pop('marksize')
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'line'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_line
        return self
    
    def _execute_plot_line(self, fig, ax, i, j):
        ax_plot = ax.plot(j['xdata'], j['ydata'], **j['kwargs'])
    


class Bar():
    def add_bar(self, name, xdata, ydata, **kwargs):
        """Make a bar plot.
        
        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
            width : float or array-like, default: 0.8
                The width(s) of the bars.

            vertical: bool, default: True
            
            bottom : float or array-like, default: 0
                The y coordinate(s) of the bars bases.

            align : {'center', 'edge'}, default: 'center'
                Alignment of the bars to the *x* coordinates:

                - 'center': Center the base on the *x* positions.
                - 'edge': Align the left edges of the bars with the *x* positions.
                
            barcolor : color or list of color, optional
                The colors of the bar faces.

            edgecolor : color or list of color, optional
                The colors of the bar edges.

            linewidth : float or array-like, optional
                Width of the bar edge(s). If 0, don't draw edges.

            tick_label : str or list of str, optional
                The tick labels of the bars.
                Default: None (Use default numeric labels.)

            xerr, yerr : float or array-like of shape(N,) or shape(2, N), optional
                If not *None*, add horizontal / vertical errorbars to the bar tips.
                The values are +/- sizes relative to the data:

                - scalar: symmetric +/- values for all bars
                - shape(N,): symmetric +/- values for each bar
                - shape(2, N): Separate - and + values for each bar. First row
                  contains the lower errors, the second row contains the upper
                  errors.
                - *None*: No errorbar. (Default)

                See :doc:`/gallery/statistics/errorbar_features`
                for an example on the usage of ``xerr`` and ``yerr``.

            ecolor : color or list of color, default: 'black'
                The line color of the errorbars.

            capsize : float, default: :rc:`errorbar.capsize`
               The length of the error bar caps in points.

            error_kw : dict, optional
                Dictionary of kwargs to be passed to the `~.Axes.errorbar`
                method. Values of *ecolor* or *capsize* defined here take
                precedence over the independent kwargs.

            log : bool, default: False
                If *True*, set the y-axis to be log scale.

            linestyle: line style, {'-', '--', '-.', ':'}.
                '-' or 'solid': solid line
                '--' or 'dashed': dashed line
                '-.' or 'dashdot': dash-dotted line
                ':' or 'dotted': dotted line
                'none', 'None', ' ', or '': draw nothing
            
            labels: array-like, optional
                A list of label texts, that should be displayed. 
                If not given, the label texts will be the data values formatted with fmt.

            label_type: {'edge', 'center'}, default: 'edge'
                The label type. Possible values:
                'edge': label placed at the end-point of the bar segment, 
                        and the value displayed will be the position of that end-point.
                'center': label placed in the center of the bar segment, 
                and the value displayed will be the length of that segment. 

            padding: float, default: 0
                Distance of label from the end of the bar, in points.
        """
        if 'barcolor' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]
        elif isinstance(kwargs['barcolor'], dict):
            kwargs['color'] = kwargs.pop('barcolor')['mode']
        else:
            kwargs['color'] = kwargs.pop('barcolor')
        
        barlabel = {}
        if 'labels' in kwargs:
            barlabel['labels'] = kwargs.pop('labels')
        if 'label_type'in kwargs:
            barlabel['label_type'] = kwargs.pop('label_type')
        if 'padding' in kwargs:
            barlabel['padding'] = kwargs.pop('padding')
        
        self._params.ydata[name]['vertical'] = kwargs.pop('vertical') if 'vertical' in kwargs else True
        if not self._params.ydata[name]['vertical']:
            if 'width' in kwargs:
                kwargs['height'] = kwargs.pop('width')
            if 'bottom' in kwargs:
                kwargs['left'] = kwargs.pop('bottom')
        self._params.ydata[name]['barlabel'] = barlabel
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'bar'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_bar
        return self
    
    def _execute_plot_bar(self, fig, ax, i, j):
        if j['vertical']:
            ax_plot = ax.bar(j['xdata'], j['ydata'], **j['kwargs'])
        else:
            ax_plot = ax.barh(j['xdata'], j['ydata'], **j['kwargs'])
        ax_plot.set_label(i)
        if len(j['barlabel'])>0:
            if 'label_type' in j['barlabel']:
                if isinstance(j['barlabel']['label_type'], str):
                    label_type = [j['barlabel']['label_type']]
                else:
                    label_type = j['barlabel']['label_type']
            else:
                label_type = ['edge']
            t = j['barlabel'].copy()
            for i in label_type:
                t['label_type'] = i
                ax.bar_label(ax_plot, **t)


from linora.utils._config import Config

__all__ = ['ModelCheckpoint']


class ModelCheckpoint():
    """Callback to save model at some frequency.
    
    Args:
        monitor: Quantity to be monitored.
        patience: The number of batches for the training monitoring interval.
        mode: One of {"min", "max"}. the current save file is made based on either the 
            maximization or the minimization of the monitored quantity.
    """
    def __init__(self, monitor, patience=10, mode='min'):
        self._params = Config()
        self._params.monitor = monitor
        self._params.patience = patience
        self._params.mode = mode
        self._params.history = []
        self._params.checkpoint = False
        self._params.polt_num = 0
        self._params.best = None
        self._params.name = 'ModelCheckpoint'
        
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.monitor in log:
            self._params.history += [log[self._params.monitor]]
            if self._params.polt_num%self._params.patience==0:
                if self._params.best is None:
                    self._params.checkpoint = True
                    self._params.best = self._params.history[0]
                elif self._params.mode=='min':
                    self._params.checkpoint = min(self._params.history)<self._params.best
                else:
                    self._params.checkpoint = max(self._params.history)>self._params.best
                self._params.history = []
            else:
                self._params.checkpoint = False
            self._params.polt_num += 1


import json
from collections import defaultdict

import numpy as np
import matplotlib.pyplot as plt
from IPython.display import clear_output

from linora.chart._config import Options
from linora.utils._config import Config

__all__ = ['Visual']


class Visual():
    """Real-time dynamic training visualization callback.
    
    Args:
        ncols : int, default 2, The number of sub graphs that the width of metrics
                 visualiztion image to accommodate at most;
        batch_total : int, default None, Pre-specify the maximum value of x-axis in each
                  sub-picture to indicate the maximum number of batch or epoch training;
        batch_wait : int, default 5, Indicates how many batches or epochs are drawn
                  each time a graph is drawn;
        batch_max_draw: How many data to draw.
        avg_num: int, default 1, mertics move aveage.
        mode : int, default 1, 1 means the x-axis name is 'batch', 0 means the x-axis name is 'epoch';
        figsize : tuple, default None，Represents the customize image size;
        valid_fmt : str, default "val_{}",The string preceding the underscore is used to
                   instruction the training and validation is displayed together in the
                   same sub graph. The training indicator is not required to have a prefix.
                   The validation indicator prefix is 'val' in the "val_{}";
    """
    def __init__(self, ncols=2, batch_total=None, batch_wait=20, batch_max_draw=100, avg_num=5,
                 mode=1, figsize=None, valid_fmt="test_{}"):
        self._params = Config()
        self._params.ncols = ncols
        self._params.mode = mode
        self._params.batch_total = batch_total
        self._params.batch_wait = batch_wait
        self._params.batch_max_draw = batch_max_draw+1
        self._params.figsize = figsize
        self._params.valid_fmt = valid_fmt
        self._params.xlabel = {0:'epoch', 1:'batch'}
        self._params.figure = None
        self._params.avg_num = avg_num if avg_num>=0 else 1
        self._params.metrics = defaultdict()
        self.history = defaultdict(lambda: defaultdict(list))
        key = np.random.choice(list(Options.color), size=len(Options.color), replace=False)
        t = {i:np.random.choice(Options.color[i], size=len(Options.color[i]), replace=False) for i in key}
        self._params.color = sorted([[r+k*7, j, i] for r, i in enumerate(key) for k, j in enumerate(t[i])])
        self._params.name = 'Visual'

    def _update(self, batch, log):
        """update log
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        for i in list(filter(lambda x: self._params.valid_fmt.split('_')[0] not in x.lower(), log)):
            if i not in self._params.metrics:
                self._params.metrics[i] = {'batch_total':self._params.batch_total, 
                                           'batch_max_draw':self._params.batch_max_draw, 'avg_num':self._params.avg_num}
        for metric in log:
            self.history[metric]['values'] += [log[metric]]
            self.history[metric]['index'] += [batch]            
        self._params.batch = batch
        
    def set_draw(self, monitor, batch_total=None, batch_max_draw=100, avg_num=5):
        """Metric configuration to monitor.
        
        Args:
            monitor: Metric to monitor.
            batch_total : int, default None, Pre-specify the maximum value of x-axis in each
                      sub-picture to indicate the maximum number of batch or epoch training;
            batch_max_draw: How many data to draw.
            avg_num: int, default 1, mertics move aveage.
        """
        self._params.metrics[monitor] = {'batch_total':batch_total, 
                                         'batch_max_draw':batch_max_draw, 'avg_num':avg_num if avg_num>=0 else 1}
        
    def draw(self):
        """plot metrics."""
        if self._params.batch%self._params.batch_wait==0:
            clear_output(wait=True)
            self._excute()
            plt.show()
    
    def save_image(self, image_path, **kwargs):
        """save plot.
        
        Args:
            image_path: str, train end save last image.
        """
        if self._params.figure is None:
            self._excute()
        self._params.figure.savefig(image_path, **kwargs)

    def save_logs(self, filename):
        """save logs.
        
        Args:
            filename: Filename of the json file, e.g. 'run/log.json'.
        """
        with open(filename, 'w') as f:
            json.dump(self.history, f)
    
    def _excute(self):
        with plt.style.context('ggplot'):
            if self._params.figsize is None:
                figsize = (self._params.ncols*6, ((len(self._params.metrics)+1)//self._params.ncols+1)*4)
                self._params.figure = plt.figure(figsize=figsize)
            else:
                self._params.figure = plt.figure(figsize=self._params.figsize)
            for metric_id, (metric, params) in enumerate(self._params.metrics.items()):
                if self._params.batch>params['avg_num']:
                    value = np.convolve(self.history[metric]['values'], np.ones(params['avg_num']), mode='valid')/params['avg_num']
                    index = np.array(self.history[metric]['index'][-len(value)-2:])
                else:
                    value = np.array(self.history[metric]['values'])
                    index = np.array(self.history[metric]['index'])
                if len(value)>params['batch_max_draw']:
                    idx = np.linspace(0., len(value)-1, params['batch_max_draw'], dtype=int)[1:]
                else:
                    idx = np.arange(len(value))
                plt.subplot((len(self._params.metrics)+1)//self._params.ncols+1, self._params.ncols, metric_id+1)
                if params['batch_total'] is not None:
                    plt.xlim(1, params['batch_total'])
                plt.plot(index[idx], value[idx], label="train", color=self._params.color[metric_id*2][1])
                if self._params.valid_fmt.format(metric) in self.history:
                    if self._params.batch>params['avg_num']:
                        value = np.convolve(self.history[self._params.valid_fmt.format(metric)]['values'], np.ones(params['avg_num']), mode='valid')/params['avg_num']
                        index = np.array(self.history[self._params.valid_fmt.format(metric)]['index'][-len(value)-2:])
                    else:
                        value = np.array(self.history[self._params.valid_fmt.format(metric)]['values'])
                        index = np.array(self.history[self._params.valid_fmt.format(metric)]['index'])
                    if len(value)>self._params.batch_max_draw:
                        idx = np.linspace(0., len(value)-1, params['batch_max_draw'], dtype=int)[1:]
                    else:
                        idx = np.arange(len(value))
                    plt.plot(index[idx], value[idx], label=self._params.valid_fmt.split('_')[0], color=self._params.color[metric_id*2+1][1])
                plt.title(metric)
                plt.xlabel(self._params.xlabel[self._params.mode])
                plt.legend(loc='best')
                plt.axis('tight')
        plt.tight_layout()
        


from matplotlib import patches


class Arrow():
    def add_arrow(self, name, x, y, dx, dy, width=1.0, **kwargs):
        """An arrow patch.
        
        Draws an arrow from (*x*, *y*) to (*x* + *dx*, *y* + *dy*).
        The width of the arrow is scaled by *width*.
        
        Args:
            x : float, x coordinate of the arrow tail.
            y : float, y coordinate of the arrow tail.
            dx : float, Arrow length in the x direction.
            dy : float, Arrow length in the y direction.
            width : float, default: 1
                Scale factor for the width of the arrow. With a default value of 1,
                the tail width is 0.2 and head width is 0.6.
            alpha: scalar or None
            animated: bool
            antialiased or aa: unknown
            capstyle: `.CapStyle` or {'butt', 'projecting', 'round'}
            clip_box: `.Bbox`
            clip_on: bool
            clip_path: Patch or (Path, Transform) or None
            color: color
            contains: unknown
            edgecolor or ec: color or None or 'auto'
            facecolor or fc: color or None
            fill: bool
            gid: str
            hatch: {'/', '\\', '|', '-', '+', 'x', 'o', 'O', '.', '*'}
            in_layout: bool
            joinstyle: `.JoinStyle` or {'miter', 'round', 'bevel'}
            label: object
            linestyle or ls: {'-', '--', '-.', ':', '', (offset, on-off-seq), ...}
            linewidth or lw: float or None
            picker: None or bool or float or callable
            rasterized: bool
            sketch_params: (scale: float, length: float, randomness: float)
            snap: bool or None
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]
        kwargs['dx'] = dx
        kwargs['dy'] = dy
        kwargs['width'] = width
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = x
        self._params.ydata[name]['ydata'] = y
        self._params.ydata[name]['transform'] = 'ax'
        self._params.ydata[name]['plotmode'] = 'arrow'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_arrow
        return self
    
    def _execute_plot_arrow(self, fig, ax, i, j):
        poly = patches.Arrow(j['xdata'], j['ydata'], **j['kwargs'])
        if j['transform']=='ax':
            ax.add_patch(poly)
        else:
            fig.add_artist(poly)


import numpy as np


class Fillline():
    def add_fillline(self, name, xdata, ydata, ydata2=0, **kwargs):
        """A scatter plot of *y* vs. *x* with varying marker size and/or color.
        
        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
            ydata2: y-axis data.
            where : array of bool (length N), optional
                Define *where* to exclude some horizontal regions from being filled.
                The filled regions are defined by the coordinates ``x[where]``.
                More precisely, fill between ``x[i]`` and ``x[i+1]`` if
                ``where[i] and where[i+1]``.  Note that this definition implies
                that an isolated *True* value between two *False* values in *where*
                will not result in filling.  Both sides of the *True* position
                remain unfilled due to the adjacent *False* values.

            interpolate : bool, default: False
                This option is only relevant if *where* is used and the two curves
                are crossing each other.

                Semantically, *where* is often used for *y1* > *y2* or
                similar.  By default, the nodes of the polygon defining the filled
                region will only be placed at the positions in the *x* array.
                Such a polygon cannot describe the above semantics close to the
                intersection.  The x-sections containing the intersection are
                simply clipped.

                Setting *interpolate* to *True* will calculate the actual
                intersection point and extend the filled region up to this point.

            step : {'pre', 'post', 'mid'}, optional
                Define *step* if the filling should be a step function,
                i.e. constant in between *x*.  The value determines where the
                step will occur:

                - 'pre': The y value is continued constantly to the left from
                  every *x* position, i.e. the interval ``(x[i-1], x[i]]`` has the
                  value ``y[i]``.
                - 'post': The y value is continued constantly to the right from
                  every *x* position, i.e. the interval ``[x[i], x[i+1])`` has the
                  value ``y[i]``.
                - 'mid': Steps occur half-way between the *x* positions.
    
            linestyle: line style, {'-', '--', '-.', ':'}.
                       '-' or 'solid': solid line
                       '--' or 'dashed': dashed line
                       '-.' or 'dashdot': dash-dotted line
                       ':' or 'dotted': dotted line
                       'none', 'None', ' ', or '': draw nothing
            linecolor: line color, eg. 'blue' or '0.75' or 'g' or '#FFDD44' or (1.0,0.2,0.3) or 'chartreuse'.
            linewidth: line width.
            alpha: float, default: None, The alpha blending value, between 0 (transparent) and 1 (opaque).
        """
        if 'linecolor' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]#tuple([round(np.random.uniform(0, 1),1) for _ in range(3)])
        elif isinstance(kwargs['linecolor'], dict):
            kwargs['color'] = kwargs.pop('linecolor')['mode']
        else:
            kwargs['color'] = kwargs.pop('linecolor')
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['ydata2'] = ydata2
        self._params.ydata[name]['plotmode'] = 'fillline'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_fillline
        return self
    
    def _execute_plot_fillline(self, fig, ax, i, j):
        ax_plot = ax.fill_between(j['xdata'], j['ydata'], y2=j['ydata2'], **j['kwargs'])
        ax_plot.set_label(i)